import pytest
import torch

from cellseg_models_pytorch.utils import filter2D, gaussian_kernel2d, sobel_hv


@pytest.mark.parametrize("window_size", [5, 6])
def test_gaussian_kernel2d(window_size):
    kernel = gaussian_kernel2d(window_size, 2)
    assert kernel.shape == torch.Size([1, 1, window_size, window_size])


@pytest.mark.parametrize("window_size", [5, pytest.param(6, marks=pytest.mark.xfail)])
def test_sobel_kernel(window_size):
    kernel = sobel_hv(window_size)
    assert kernel.shape == torch.Size([2, 1, window_size, window_size])


def test_convolve():
    im = torch.rand([1, 1, 14, 14])
    convolved = filter2D(im, gaussian_kernel2d(5, 2))
    assert convolved.shape == im.shape



import pytest

from cellseg_models_pytorch.utils import run_pool


def wrap1(num):
    return num + 2


def wrap2(arg):
    return


@pytest.mark.parametrize(
    "typesets",
    [
        ("thread", "imap"),
        ("thread", "map"),
        ("thread", "uimap"),
        ("thread", "amap"),
        ("process", "map"),
        ("process", "amap"),
        ("process", "imap"),
        ("process", "uimap"),
        ("serial", "map"),
        ("serial", "imap"),
        pytest.param(("serial", "amap"), marks=pytest.mark.xfail),
        pytest.param(("serial", "uimap"), marks=pytest.mark.xfail),
    ],
)
@pytest.mark.parametrize("func", [wrap1, wrap2])
@pytest.mark.parametrize("ret", [True, False])
def test_run_pool(typesets, func, ret):
    args = [1, 2, 3, 4, 5, 6, 7, 8]
    res = run_pool(func, args, ret=ret, pooltype=typesets[0], maptype=typesets[1])

    if ret and func == wrap1:
        assert res == [3, 4, 5, 6, 7, 8, 9, 10]
    elif ret and func == wrap2:
        assert res == [None, None, None, None, None, None, None, None]
    else:
        # with amap, we always get a return val..
        if not ret and typesets[1] == "amap":
            pass
        else:
            assert res is None



from cellseg_models_pytorch.utils import FileHandler


def test_filehandler_get_gson(inst_map, type_map) -> None:
    split_dict = FileHandler.get_offset("x-5000_y-3000")
    assert "x" in split_dict.keys() and "y" in split_dict.keys()
    assert 3000 in split_dict.values() and 5000 in split_dict.values()

    gson = FileHandler.get_gson(
        inst_map, type_map, classes={"a": 0, "b": 1, "c": 2, "d": 3}
    )
    assert len(gson) == 7


def test_filehandler_read_h5(hdf5db):
    arrs = FileHandler.read_h5_patch(hdf5db, ix=1, return_type=True, return_sem=True)

    assert arrs["image"].any()



from pathlib import Path
from typing import Any, BinaryIO, Callable, Dict, List, Optional, Tuple

import numpy as np
import tables as tb
from tqdm import tqdm

from ...utils import FileHandler
from ._base_writer import BaseWriter

__all__ = ["HDF5Writer"]


class HDF5Writer(BaseWriter):
    def __init__(
        self,
        in_dir_im: str,
        save_dir: str,
        file_name: str,
        in_dir_mask: Optional[str] = None,
        patch_size: Optional[Tuple[int, int]] = None,
        stride: Optional[int] = None,
        transforms: Optional[List[str]] = None,
        chunk_size: int = 1,
        complevel: int = 5,
        complib: str = "blosc:lz4",
    ) -> None:
        """Write overlapping patches to a hdf5 database.

        NOTE: Very ad-hoc and not tested so there exists a chance of failure...

        Parameters
        ----------
            in_dir_im : str
                Path to the folder of images
            in_dir_mask : str
                Path to the folder of masks.
            save_dir : str
                Path to the folder where the db will be saved.
            file_name : str
                Name of the h5 db-file.
            patch_size : Tuple[int, int]
                Height and width of the extracted patches.
            stride : int
                Stride for the sliding window.
            transforms : List[str]:
                A list of transforms to apply to the images and masks. Allowed ones:
                "blur", "non_spatial", "non_rigid", "rigid", "hue_sat", "random_crop",
                "center_crop", "resize"
            chunk_size : int, default=1
                The chunk size of the h5 arrays.
            complevel : int, default=5
                Specifies a compression level for data. The allowed range is 0-9.
                A value of 0 (the default) disables compression. (PyTables).
            complib : str, default="blosc:lz4"
                Specifies the compression library to be used. 'zlib', 'lzo', 'bzip2' and
                'blosc' are supported. Additional compressors for Blosc: 'blosc:blosclz'
                'blosc:lz4', 'blosc:lz4hc', 'blosc:snappy', 'blosc:zlib' & 'blosc:zstd'.
                (PyTables).

        Raises
        ------
            ValueError if issues with given paths or the files in those folders.

        Example
        -------
            >>> # Patch and write to hdf5 db.
            >>> writer = HDF5Writer(
                    "/path/to/my/imgs/,
                    "/path/to/my/masks/,
                    "/save/db/here/,
                    "my_h5_dataset.h5",
                    patch_size=(320, 320),
                    stride=160,
                    transforms=["rigid"]
                )
            >>> writer.write(tiling=True)

            >>> # Don't patch, just write to hdf5 db.
            >>> writer = HDF5Writer(
                    "/path/to/my/imgs/,
                    "/path/to/my/masks/,
                    "/save/db/here/,
                    "my_h5_dataset.h5",
                    transforms=["rigid"]
                )
            >>> writer.write(tiling=False)
        """
        super().__init__(
            in_dir_im=in_dir_im,
            in_dir_mask=in_dir_mask,
            stride=stride,
            patch_size=patch_size,
            transforms=transforms,
        )

        self.chunk_size = chunk_size
        self.complevel = complevel
        self.complib = complib
        self.save_dir = Path(save_dir)
        self.fname = self.save_dir / file_name
        if self.fname.suffix:
            if self.fname.suffix not in (".h5", ".hdf5"):
                raise ValueError(
                    "`file_name` has to have a '.h5' or '.hdf5' suffix."
                    f"Got: {self.fname.suffix}"
                )
        else:
            self.fname = self.fname.with_suffix(".h5")

    def write(
        self,
        tiling: bool = False,
        pre_proc: Callable = None,
        msg: str = None,
    ) -> None:
        """Write patches to a hdf5 db.

        Parameters
        ----------
            tiling : bool, default=False
                Apply tiling to the images before saving.
            pre_proc : Callable, optional
                An optional pre-processing function for the masks.
        """
        h5 = tb.open_file(self.fname.as_posix(), mode="w")

        if self.patch_size is not None and tiling:
            psize = self.patch_size
        else:
            psize = FileHandler.read_img(self.fnames_im[0]).shape[:2]

        h5.create_earray(
            where=h5.root,
            name="imgs",
            atom=tb.UInt8Atom(),
            shape=np.append([0], psize + (3,)),
            chunkshape=np.append([self.chunk_size], psize + (3,)),
            filters=tb.Filters(complevel=self.complevel, complib=self.complib),
        )
        h5.create_earray(
            where=h5.root,
            name="fnames",
            atom=tb.StringAtom(itemsize=256),
            shape=(0,),
            filters=tb.Filters(complevel=self.complevel, complib=self.complib),
        )

        if self.fnames_mask is not None:
            h5.create_earray(
                where=h5.root,
                name="insts",
                atom=tb.Int32Atom(),
                shape=np.append([0], psize),
                chunkshape=np.append([self.chunk_size], psize),
                filters=tb.Filters(complevel=self.complevel, complib=self.complib),
            )

            h5.create_earray(
                where=h5.root,
                name="types",
                atom=tb.Int32Atom(),
                shape=np.append([0], psize),
                chunkshape=np.append([self.chunk_size], psize),
                filters=tb.Filters(complevel=self.complevel, complib=self.complib),
            )

            h5.create_earray(
                where=h5.root,
                name="areas",
                atom=tb.Int32Atom(),
                shape=np.append([0], psize),
                chunkshape=np.append([self.chunk_size], psize),
                filters=tb.Filters(complevel=self.complevel, complib=self.complib),
            )

        it = self.fnames_im
        if self.fnames_mask is not None:
            it = zip(self.fnames_im, self.fnames_mask)

        try:
            with tqdm(it, total=len(self.fnames_im)) as pbar:

                total_tiles = 0
                msg = msg if msg is not None else ""
                for fn in pbar:

                    # get the img and masks filenames
                    if self.fnames_mask is not None:
                        fn_im, fn_mask = fn
                    else:
                        fn_im = fn
                        fn_mask = None

                    pbar.set_description(f"Extracting {msg} patches to h5 db..")

                    # optionally patch and process images and masks
                    im, masks = self.get_array(
                        fn_im, fn_mask, tiling=tiling, pre_proc=pre_proc
                    )

                    n_tiles = im.shape[0] if tiling else 1

                    arg_list = []
                    for i in range(n_tiles):
                        kw = self._set_kwargs(im, fn_im, masks, fn_mask, tiling, i)
                        kw["h5"] = h5
                        arg_list.append(kw)

                    self._write_sequential(self.save2db, arg_list)
                    total_tiles += n_tiles
                    pbar.set_postfix_str(f"# of extracted tiles {total_tiles}")
        except Exception as e:
            h5.close()
            print(e)

        h5.root._v_attrs.n_items = h5.root.imgs.shape[0]
        h5.close()

    def _set_kwargs(
        self,
        im: np.ndarray,
        img_path: Path,
        masks: Dict[str, np.ndarray] = None,
        mask_path: Path = None,
        tiling: bool = False,
        ix: int = None,
    ) -> Dict[str, Any]:
        """Define one set of arguments for saving function."""
        fn = img_path.with_suffix("").name

        fn_im_new = fn
        fn_mask_new = None
        if mask_path is not None:
            fn_mask_new = fn

        if tiling:
            fn_im_new = f"{fn}_patch{ix + 1}"

            if mask_path is not None:
                fn_mask_new = f"{fn}_patch{ix + 1}"

        im_save_path = self.save_dir / fn_im_new

        mask_save_path = None
        if mask_path is not None:
            mask_save_path = self.save_dir / fn_mask_new

        kwarg = {"path_im": im_save_path, "path_mask": mask_save_path}

        if tiling:
            kwarg["image"] = im[ix : ix + 1]

            if mask_path is not None:
                for k, m in masks.items():
                    kwarg[k] = m[ix : ix + 1]
        else:
            kwarg["image"] = im[None, ...]
            if mask_path is not None:
                for k, m in masks.items():
                    kwarg[k] = m[None, ...]

        return kwarg

    def save2db(
        self,
        h5: BinaryIO,
        path_im: str,
        image: np.ndarray,
        path_mask: str = None,
        inst_map: np.ndarray = None,
        type_map: np.ndarray = None,
        sem_map: np.ndarray = None,
    ) -> None:
        """Write image and corresponding masks to h5-db."""
        h5.root.imgs.append(image)
        h5.root.fnames.append(np.array([path_im], dtype=str))

        if path_mask is not None:
            h5.root.insts.append(inst_map)

            if type_map is not None:
                h5.root.types.append(type_map)
            if sem_map is not None:
                h5.root.areas.append(sem_map)



from typing import List

import albumentations as A
import cv2
from albumentations.core.transforms_interface import BasicTransform

__all__ = [
    "rigid_transforms",
    "non_rigid_transforms",
    "non_spatial_transforms",
    "hue_saturation_transforms",
    "blur_transforms",
    "random_crop",
    "center_crop",
    "resize",
]


def rigid_transforms(**kwargs) -> List[BasicTransform]:
    """Return rigid albumentations augmentations.

    For every patch, either:
    - Rotate
    - random rotate 90 degrees
    - flip (rotate 180 degrees)
    - transpose (flip x and y axis)
    is applied with a probability of 0.5*(0.5/(0.5+0.5+0.5+0.5))=0.125

    Returns
    -------
        List[BasicTransform]:
            A List of possible albumentations data augmentations.
    """
    return [
        A.OneOf(
            [
                A.Rotate(p=0.8),
                A.RandomRotate90(p=0.5),
                A.Flip(p=0.5),
                A.Transpose(p=0.5),
            ],
            p=0.5,
        )
    ]


def non_rigid_transforms(**kwargs) -> List[BasicTransform]:
    """Return non rigid albumentations augmentations.

    For every patch, either:
    - elastic transformation
    - grid distortion
    - optical distortion
    is applied with a probability of 0.3*(0.5/(0.5+0.5+0.5))=0.1

    Returns
    -------
        List[BasicTransform]:
            A List of possible albumentations data augmentations.
    """
    return [
        A.OneOf(
            [
                A.ElasticTransform(
                    alpha=120, sigma=120 * 0.05, alpha_affine=120 * 0.03
                ),
                A.GridDistortion(p=0.5),
                A.OpticalDistortion(distort_limit=2, shift_limit=0.5, p=0.5),
            ],
            p=0.3,
        )
    ]


def hue_saturation_transforms(**kwargs) -> List[BasicTransform]:
    """Return hue saturation albumentations augmentations.

    For every patch, either:
    - hue saturation value shift is applied with a probability of 0.5*0.25*0.125

    Returns
    -------
        List[BasicTransform]:
            A List of possible albumentations data augmentations.
    """
    return [
        A.OneOf(
            [
                A.HueSaturationValue(
                    hue_shift_limit=(0, 15),
                    sat_shift_limit=(0, 30),
                    val_shift_limit=(0, 20),
                    p=0.25,
                )
            ],
            p=0.5,
        )
    ]


def blur_transforms(**kwargs) -> List[BasicTransform]:
    """Return blur albumentations augmentations.

    For every patch, either:
    - motion blur
    - median blur
    - gaussian blur
    is applied with a probability of 0.3*(0.5/(0.5+0.5+0.5))=0.1

    Returns
    -------
        List[BasicTransform]:
            A List of possible albumentations data augmentations.
    """
    return [
        A.OneOf(
            [
                A.MotionBlur(blur_limit=7, p=0.5),
                A.MedianBlur(blur_limit=7, p=0.5),
                A.Blur(blur_limit=7, p=0.5),
            ],
            p=0.3,
        )
    ]


def non_spatial_transforms(**kwargs) -> List[BasicTransform]:
    """Return non spatial albumentations augmentations.

    For every patch, either:
    - CLAHE
    - brightness contrast
    - random gamma
    is applied with a probability of 0.7*(0.5/(0.5+0.5+0.5))=0.233

    Returns
    -------
        List[BasicTransform]:
            A List of possible albumentations data augmentations.
    """
    return [
        A.OneOf(
            [A.CLAHE(p=0.5), A.RandomBrightnessContrast(p=0.5), A.RandomGamma(p=0.5)],
            p=0.5,
        )
    ]


def center_crop(height: int, width: int, **kwargs) -> List[BasicTransform]:
    """Return albumentations center crop.

    For every patch a crop a is extracted from the center, p=1.0.

    Parameters
    ----------
        height : int
            Height of the output image.
        width : int
            Width of the input image.

    Returns
    -------
        List[BasicTransform]:
            A List containing the center crop transform.
    """
    return [
        A.OneOf(
            [A.CenterCrop(height=height, width=width, always_apply=True, p=1)], p=1.0
        )
    ]


def random_crop(height: int, width: int, **kwargs) -> List[BasicTransform]:
    """Return albumentations random crop.

    For every patch a crop a is extracted randomly, p=1.0.

    Parameters
    ----------
        height : int
            Height of the output image.
        width : int
            Width of the input image.

    Returns
    -------
        List[BasicTransform]:
            A List containing the random crop transform.
    """
    return [
        A.OneOf(
            [A.RandomCrop(height=height, width=width, always_apply=True, p=1)], p=1.0
        )
    ]


def resize(height: int, width: int, **kwargs) -> List[BasicTransform]:
    """Return albumentations resize transform.

    Parameters
    ----------
        height : int
            Height of the output image
        width : int
            Width of the input image

    Returns
    -------
        List[BasicTransform]:
            A List containing the resize transform.
    """
    return [
        A.OneOf(
            [A.Resize(height=height, width=width, interpolation=cv2.INTER_CUBIC, p=1)]
        )
    ]



from .binary_metrics import accuracy, dice_coef, f1score, get_stats, iou_score
from .functional import (
    accuracy_multiclass,
    aggregated_jaccard_index,
    average_precision,
    dice2,
    dice_multiclass,
    f1score_multiclass,
    iou_multiclass,
    pairwise_object_stats,
    pairwise_pixel_stats,
    panoptic_quality,
    sensitivity_multiclass,
    specificity_multiclass,
)

__all__ = [
    "get_stats",
    "accuracy",
    "iou_score",
    "f1score",
    "dice_coef",
    "pairwise_pixel_stats",
    "panoptic_quality",
    "dice2",
    "average_precision",
    "pairwise_object_stats",
    "aggregated_jaccard_index",
    "iou_multiclass",
    "dice_multiclass",
    "f1score_multiclass",
    "accuracy_multiclass",
    "sensitivity_multiclass",
    "specificity_multiclass",
]



import torch
import torch.nn as nn

__all__ = ["FixedUnpool"]


# Adapted from
# https://github.com/vqdang/hover_net/blob/master/models/hovernet/net_utils.py
class FixedUnpool(nn.Module):
    def __init__(self, scale_factor: int = 2) -> None:
        """Upsample input by a scale factor.

        TensorPack (library) fixed unpooling in pytorch.

        Parameters
        ----------
            scale_factor : int, default=2
                Upsampling scale factor. scale_factor*(H, W)
        """
        super().__init__()
        self.scale_factor = scale_factor
        self.register_buffer(
            "unpool_mat", torch.ones([scale_factor, scale_factor], dtype=torch.float32)
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Fixed-unpooling forward pass."""
        in_shape = list(x.shape)
        x = x.unsqueeze(-1)  # (B, C, H, W)
        mat = self.unpool_mat.unsqueeze(0)  # (B, C, H, W, 1)
        ret = torch.tensordot(x, mat, dims=1)  # (B, C, H, W, SH, SW)
        ret = ret.permute(0, 1, 2, 4, 3, 5)
        ret = ret.reshape(
            (
                -1,
                in_shape[1],
                in_shape[2] * self.scale_factor,
                in_shape[3] * self.scale_factor,
            )
        )

        return ret

    def extra_repr(self):
        """Print output."""
        return f"scale_factor={self.scale_factor}"  # .format(**self.__dict__)



from torch.nn import ConvTranspose2d, Upsample

from .fixed_unpool import FixedUnpool

UP_LOOKUP = {
    "fixed-unpool": FixedUnpool,
    "bilinear": Upsample,
    "bicubic": Upsample,
    "transconv": ConvTranspose2d,
}

__all__ = ["UP_LOOKUP", "FixedUnpool", "ConvTranspose2d", "Upsample"]



import pytest
import torch

from cellseg_models_pytorch.decoders import Decoder


@pytest.mark.parametrize(
    "long_skip", ["unet", "unetpp", "unet3p", "unet3p-lite", "cross-attn"]
)
@pytest.mark.parametrize("merge_policy", ["cat", "sum"])
@pytest.mark.parametrize("use_conv", [True, False])
@pytest.mark.parametrize("use_tr", [True, False])
def test_decoder_fwdbwd(long_skip, merge_policy, use_conv, use_tr):
    enc_channels = (32, 32, 32, 32)
    out_dims = [32 // 2**i for i in range(4)][::-1]

    decoder1_kwargs = {"merge_policy": merge_policy}
    decoder2_kwargs = {"merge_policy": merge_policy}
    decoder3_kwargs = {"merge_policy": merge_policy}
    decoder4_kwargs = {"merge_policy": merge_policy}
    decoder5_kwargs = {"merge_policy": merge_policy}
    stage_params = (
        decoder1_kwargs,
        decoder2_kwargs,
        decoder3_kwargs,
        decoder4_kwargs,
        decoder5_kwargs,
    )

    n_layers = None
    n_blocks = None
    if use_conv:
        n_layers = (1, 1, 1, 1)
        n_blocks = ((2,), (2,), (2,), (2,))

    n_tr_layers = None
    n_tr_blocks = None
    if use_tr:
        n_tr_layers = (1, 1, 1, 1)
        n_tr_blocks = ((1,), (1,), (1,), (1,))

    decoder = Decoder(
        enc_channels=enc_channels,
        out_channels=(32, 32, 32, 32),
        n_conv_layers=n_layers,
        n_conv_blocks=n_blocks,
        n_transformers=n_tr_layers,
        n_transformer_blocks=n_tr_blocks,
        long_skip=long_skip,
        stage_params=stage_params,
    )

    x = [torch.rand([1, enc_channels[i], out_dims[i], out_dims[i]]) for i in range(4)]
    out = decoder(*x)

    out[-1].mean().backward()

    assert out[-1].shape[1] == decoder.out_channels


@pytest.mark.slow
@pytest.mark.parametrize("long_skip", ["unet", "unetpp", "unet3p", "unet3p-lite"])
@pytest.mark.parametrize("merge_policy", ["cat", "sum"])
@pytest.mark.parametrize("short_skip", ["residual", "dense", "basic"])
@pytest.mark.parametrize(
    "block_types", ["mbconv", "fmbconv", "basic", "bottleneck", "dws", "hover_dense"]
)
@pytest.mark.parametrize("normalizations", ["bn", "bcn", "gn"])
@pytest.mark.parametrize("convolutions", ["conv", "scaled_wsconv", "wsconv"])
@pytest.mark.parametrize("attentions", ["se", "scse", "eca", "gc"])
def test_decoder_fwdbwd_all(
    long_skip,
    merge_policy,
    short_skip,
    block_types,
    normalizations,
    convolutions,
    attentions,
):
    enc_channels = (64, 32, 16, 8, 8)
    out_dims = [256 // 2**i for i in range(6)][::-1]

    decoder1_kwargs = {
        "short_skips": (short_skip,),
        "block_types": ((block_types, block_types),),
        "normalizations": ((normalizations, normalizations),),
        "activations": (("leaky-relu", "leaky-relu"),),
        "convolutions": ((convolutions, convolutions),),
        "attentions": ((attentions, attentions),),
        "merge_policy": merge_policy,
        "skip_params": {
            "convolutions": ((convolutions,),),
            "normalizations": ((normalizations,),),
            "attentions": ((attentions,),),
            "block_types": ((block_types,),),
        },
    }

    decoder2_kwargs = {
        "short_skips": (short_skip,),
        "block_types": ((block_types, block_types),),
        "normalizations": ((normalizations, normalizations),),
        "activations": (("leaky-relu", "leaky-relu"),),
        "convolutions": ((convolutions, convolutions),),
        "attentions": ((attentions, attentions),),
        "merge_policy": merge_policy,
        "skip_params": {
            "convolutions": ((convolutions,),),
            "normalizations": ((normalizations,),),
            "attentions": ((attentions,),),
            "block_types": ((block_types,),),
        },
    }

    decoder3_kwargs = {
        "short_skips": (short_skip,),
        "block_types": ((block_types, block_types),),
        "normalizations": ((normalizations, normalizations),),
        "activations": (("leaky-relu", "leaky-relu"),),
        "convolutions": ((convolutions, convolutions),),
        "attentions": ((attentions, attentions),),
        "merge_policy": merge_policy,
        "skip_params": {
            "convolutions": ((convolutions,),),
            "normalizations": ((normalizations,),),
            "attentions": ((attentions,),),
            "block_types": ((block_types,),),
        },
    }

    decoder4_kwargs = {
        "short_skips": (short_skip,),
        "block_types": ((block_types, block_types),),
        "normalizations": ((normalizations, normalizations),),
        "activations": (("leaky-relu", "leaky-relu"),),
        "convolutions": ((convolutions, convolutions),),
        "attentions": ((attentions, attentions),),
        "merge_policy": merge_policy,
        "skip_params": {
            "convolutions": ((convolutions,),),
            "normalizations": ((normalizations,),),
            "attentions": ((attentions,),),
            "block_types": ((block_types,),),
        },
    }

    decoder5_kwargs = {
        "short_skips": (short_skip,),
        "block_types": ((block_types, block_types),),
        "normalizations": ((normalizations, normalizations),),
        "activations": (("leaky-relu", "leaky-relu"),),
        "convolutions": ((convolutions, convolutions),),
        "attentions": ((attentions, attentions),),
        "merge_policy": merge_policy,
        "skip_params": {
            "convolutions": ((convolutions,),),
            "normalizations": ((normalizations,),),
            "attentions": ((attentions,),),
            "block_types": ((block_types,),),
        },
    }
    stage_params = (
        decoder1_kwargs,
        decoder2_kwargs,
        decoder3_kwargs,
        decoder4_kwargs,
        decoder5_kwargs,
    )

    decoder = Decoder(
        enc_channels=enc_channels,
        model_input_size=256,
        out_channels=(64, 32, 16, 8, 8),
        n_layers=(1, 1, 1, 1, 1),
        n_blocks=((2,), (2,), (2,), (2,), (2,)),
        long_skip=long_skip,
        stage_params=stage_params,
    )

    x = [torch.rand([1, enc_channels[i], out_dims[i], out_dims[i]]) for i in range(5)]
    out = decoder(*x)

    out[-1].mean().backward()

    assert out[-1].shape[1] == decoder.out_channels



from collections import OrderedDict

import pytest
import torch

from cellseg_models_pytorch.losses import (
    SSIM,
    CELoss,
    IoULoss,
    JointLoss,
    MultiTaskLoss,
    TverskyLoss,
)
from cellseg_models_pytorch.losses.tests.test_losses import _get_dummy_pair


@torch.no_grad()
@pytest.mark.parametrize("n_classes", [1, 3])
@pytest.mark.parametrize(
    "losses",
    [
        {"inst": JointLoss([TverskyLoss(), IoULoss()])},
        {"inst": CELoss(), "type": SSIM()},
    ],
)
def test_multitask_loss(n_classes, losses):
    losses = OrderedDict(losses)

    yhats = {}
    targets = {}
    for i, br in enumerate(losses.keys(), 1):
        yhats[f"{br}"], targets[f"{br}"] = _get_dummy_pair(n_classes + i)

    mtl = MultiTaskLoss(losses)
    mtl(yhats, targets)



from typing import Any, Dict

import torch
import torch.nn as nn
import torch.nn.functional as F
from timm.models.layers import DropPath

from .base_modules import Identity
from .misc_modules import LayerScale
from .mlp import MlpBlock
from .patch_embeddings import ContiguousEmbed
from .token_mixers import RESHAPE_LOOKUP, TokenMixerBlock


class MetaFormer(nn.Module):
    def __init__(
        self,
        in_channels: int,
        embed_kwargs: Dict[str, Any],
        mixer_kwargs: Dict[str, Any],
        mlp_kwargs: Dict[str, Any],
        out_channels: int = None,
        layer_scale: bool = False,
        dropout: float = 0.0,
        **kwargs
    ) -> None:
        """Create a generic Metaformer block with any token-mixer available.

        Input shape: (B, in_channels, H, W)
        Output shape: (B, out_channels, H, W)

        Parameters
        ----------
            in_channels : int
                Number of input channels.
            embed_kwargs : Dict[str, Any]
                Key-word arguments for the patch embedding block.
            mixer_kwargs : Dict[str, Any]
                Key-word arguments for the token-mixer block.
            mlp_kwargs : Dict[str, Any]
                Key-word arguments for the final Mlp-block.
            out_channels : int, optional
                Number of output channels.
            layer_scale : bool, default=False
                Flag, whether to use layer-scaling.
            dropout : float, default=0.0
                Drop-path probaility.

        Examples
        --------
        MetaFormer with exact memory-efficient self-attention:
        >>> import torch
        >>> import torch.nn as nn

        >>> in_channels = 3
        >>> head_dim = 64
        >>> num_heads = 8
        >>> query_dim = head_dim*num_heads

        >>> # patch embedding kwargs
        >>> embed_kwargs = {
                "in_channels": 3,
                "kernel_size": 7,
                "stride": 4,
                "pad": 2,
                "head_dim": head_dim,
                "num_heads": num_heads,
            }

        >>> # token-mixer kwargs
        >>> mixer_kwargs = {
                "token_mixer": "self-attention",
                "normalization": "ln",
                "residual": True,
                "norm_kwargs": {
                    "normalized_shape": query_dim
                },
                "mixer_kwargs": {
                    "query_dim": query_dim,
                    "name": "exact",
                    "how": "memeff",
                    "cross_attention_dim": None,
                }
            }

        >>> # mlp-kwargs
        >>> mlp_kwargs = {
                "in_channels": query_dim,
                "norm_kwargs": {"normalized_shape": query_dim}
            }

        >>> # init metaformer
        >>> metaformer = MetaFormer(
                in_channels=in_channels,
                embed_kwargs=embed_kwargs,
                mixer_kwargs=mixer_kwargs,
                mlp_kwargs=mlp_kwargs,
                layer_scale=True,
                dropout=0.1
            )

        >>> x = torch.rand([8, 3, 256, 256])
        >>> print(metaformer(x).shape)
        >>> # torch.Size([8, 4096, 512])


        MetaFormer with multi-scale convolutional attention.:
        >>> import torch
        >>> import torch.nn as nn

        >>> in_channels = 3
        >>> head_dim = 64
        >>> num_heads = 8
        >>> query_dim = head_dim*num_heads
        >>> out_channels = 128

        >>> # patch embedding kwargs
        >>> embed_kwargs = {
                "in_channels": 3,
                "kernel_size": 7,
                "stride": 4,
                "pad": 2,
                "head_dim": head_dim,
                "num_heads": num_heads,
            }

        >>> # token-mixer kwargs
        >>> mixer_kwargs = {
            "token_mixer": "mscan",
            "normalization": "bn",
            "norm_kwargs": {
                "num_features": query_dim,
            },
            "mixer_kwargs":{
                "in_channels": query_dim,
            }
        }

        >>> # mlp-kwargs
        >>> mlp_kwargs = {
                "in_channels": query_dim,
                "norm_kwargs": {"normalized_shape": query_dim}
            }

        >>> # init metaformer
        >>> metaformer = MetaFormer(
                in_channels=in_channels,
                out_channels=out_channels,
                embed_kwargs=embed_kwargs,
                mixer_kwargs=mixer_kwargs,
                mlp_kwargs=mlp_kwargs,
                layer_scale=True,
                dropout=0.1
            )

        >>> x = torch.rand([8, 3, 256, 256])
        >>> print(metaformer(x).shape)
        >>> # torch.Size([8, 128, 256, 256])
        """
        super().__init__()
        self.out_channels = out_channels if out_channels is not None else in_channels
        mixer_name = mixer_kwargs["token_mixer"]

        self.patch_embed = ContiguousEmbed(
            **embed_kwargs, flatten=not RESHAPE_LOOKUP[mixer_name]
        )
        self.proj_dim = self.patch_embed.proj_dim

        self.mixer = TokenMixerBlock(**mixer_kwargs)
        self.mlp = MlpBlock(**mlp_kwargs)
        self.ls1 = (
            LayerScale(dim=mlp_kwargs["in_channels"]) if layer_scale else Identity()
        )
        self.ls2 = (
            LayerScale(dim=mlp_kwargs["in_channels"]) if layer_scale else Identity()
        )
        self.drop_path1 = DropPath() if dropout else Identity()
        self.drop_path2 = DropPath() if dropout else Identity()

        self.proj_out = nn.Conv2d(
            self.proj_dim, self.out_channels, kernel_size=1, stride=1, padding=0
        )

        self.downsample = Identity()
        if self.out_channels is not None:
            self.downsample = nn.Conv2d(in_channels, out_channels, 1)

    def forward(self, x: torch.Tensor, **kwargs) -> torch.Tensor:
        """Forward pass of the token mixer module."""
        B, _, H, W = x.shape
        residual = self.downsample(x)

        # 1. embed and project
        x = self.patch_embed(x)

        # 2. token-mixing
        x = self.drop_path1(self.ls1(self.mixer(x, **kwargs)))

        # 3. mlp
        x = self.drop_path2(self.ls2(self.mlp(x)))

        # 4. Reshape back to image-like shape.
        p_H = self.patch_embed.get_patch_size(H)
        p_W = self.patch_embed.get_patch_size(W)
        x = x.reshape(B, p_H, p_W, self.proj_dim).permute(0, 3, 1, 2)

        # Upsample to input dims if patch size less than orig inp size
        # assumes that the input is square mat.
        # NOTE: the kernel_size, pad, & stride has to be set correctly for this to work
        if p_H < H:
            scale_factor = H // p_H
            x = F.interpolate(x, scale_factor=int(scale_factor), mode="bilinear")

        # 5. project to original input channels
        x = self.proj_out(x)

        # 6. residual
        return x + residual



from typing import Tuple

import torch
import torch.nn as nn

from .base_modules import MultiHeadSelfAttention

__all__ = ["SelfAttention", "SelfAttentionBlock"]


class SelfAttention(nn.Module):
    def __init__(
        self,
        query_dim: int,
        name: str = "exact",
        how: str = "basic",
        cross_attention_dim: int = None,
        num_heads: int = 8,
        head_dim: int = 64,
        dropout: float = 0.0,
        bias: bool = False,
        slice_size: int = 4,
        **kwargs,
    ) -> None:
        """Compute self-attention.

        Includes all the data wrangling on the input before attention computation.

        Input Shape: (B, H'*W', query_dim).
        Output Shape: (B, H'*W', query_dim).

        Parameters
        ----------
            query_dim : int
                The number of channels in the query. Typically: num_heads*head_dim
            name : str
                Name of the attention method. One of ("exact", "linformer").
            how : str, default="basic"
                How to compute the self-attention matrix.
                One of ("basic", "flash", "slice", "memeff", "slice-memeff").
                "basic": the normal O(N^2) self attention.
                "flash": the flash attention (by xformers library),
                "slice": batch sliced attention operation to save mem.
                "memeff": xformers.memory_efficient_attention.
                "slice-memeff": Conmbine slicing and memory_efficient_attention.
            cross_attention_dim : int, optional
                Number of channels in the context tensor. Cross attention combines
                asymmetrically two separate embeddings (context and input embeddings).
                E.g. passage from transformer encoder to transformer decoder. If this is
                set to None, no cross attention is applied.
            num_heads : int, default=8
                Number of heads for multi-head attention.
            head_dim : int, default=64
                Number of output channels per head.
            dropout : float, default=0.0
                Dropout probability.
            bias : bool, default=False
                Flag to set bias for Q, K and V.
            slice_size : int, default=4
                Slice size for sliced self-attention. This is used only if
                `self_attention = "slice"`.
            **kwargs:
                Extra key-word arguments for the MHSA-module
        """
        super().__init__()
        self.out_channels = query_dim
        self.num_heads = num_heads
        proj_dim = head_dim * num_heads

        # cross attention dim
        if cross_attention_dim is None:
            cross_attention_dim = query_dim

        self.to_q = nn.Linear(query_dim, proj_dim, bias=bias)
        self.to_k = nn.Linear(cross_attention_dim, proj_dim, bias=bias)
        self.to_v = nn.Linear(cross_attention_dim, proj_dim, bias=bias)

        self.self_attn = MultiHeadSelfAttention(
            name=name,
            head_dim=head_dim,
            num_heads=self.num_heads,
            how=how,
            slice_size=slice_size,
            **kwargs,
        )

        self.to_out = nn.Linear(proj_dim, query_dim)
        self.dropout = nn.Dropout(dropout) if bool(dropout) else None

    def to_qkv(
        self, features: torch.Tensor, context: torch.Tensor = None
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Get query, key and value tensors.

        NOTE: If a context tensor is given, the key and value will be computed
        from that. If `context=None`, key and value is computed from the input
        features. This allows cross-attention.

        Parameters
        ----------
            features : torch.Tensor
                Input tensor. Usually a projection of the input features into the shape
                of a query tensor. Shape: (B, H*W, proj_dim).
            context : torch.Tensor, optional
                A context tensor. Same shape as `features`.

        Returns
        -------
            Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
                The query, key & value tensors.
                Shaped: (B*num_heads, H*W, proj_dim//num_heads).
        """
        query = self.to_q(features)
        context = context if context is not None else features
        key = self.to_k(context)
        value = self.to_v(context)

        query = self._heads2batch(query)
        key = self._heads2batch(key)
        value = self._heads2batch(value)

        return query, key, value

    def _heads2batch(self, x: torch.Tensor) -> torch.Tensor:
        """Reshape `x` input for self-attention.

        NOTE:
        Step 1. Divide the `proj_dim` (num_heads*head_out_channels) long projection
        vectors of pixel vals into `num_heads` (default=8) number of vectors that are
        `ìnner_dim`//`num_heads` (default=64) long.

        Step 2. Then bake the heads dim into batch dimension.

        Parameters
        ----------
            x : torch.Tensor
                Input tensor. Either, the query, key or value tensor.
                Shape: (B, H*W, proj_dim).

        Returns
        -------
            torch.Tensor
                A reshaped input of shape: (B*num_heads, H*W, proj_dim//num_heads).
        """
        B, seq_len, proj_dim = x.shape
        x = x.reshape(B, seq_len, self.num_heads, proj_dim // self.num_heads)
        x = x.permute(0, 2, 1, 3).reshape(
            B * self.num_heads, seq_len, proj_dim // self.num_heads
        )

        return x

    def _batch2heads(self, x: torch.Tensor) -> torch.Tensor:
        """Reshape the attention scores back to (B, H*W, proj_dim).

        Parameters
        ----------
            x : torch.Tensor
                Input tensor. The attention scores from the self attention computation.
                Shape: (B*num_heads, H*W, proj_dim//num_heads).

        Returns
        -------
            torch.Tensor:
                A reshaped attention score tensor of shape: (B, H*W, proj_dim).
        """
        # num_heads = self.num_heads
        B, seq_len, head_out_channels = x.shape
        x = x.reshape(B // self.num_heads, self.num_heads, seq_len, head_out_channels)
        x = x.permute(0, 2, 1, 3).reshape(
            B // self.num_heads, seq_len, head_out_channels * self.num_heads
        )

        return x

    def forward(
        self, features: torch.Tensor, context: torch.Tensor = None
    ) -> torch.Tensor:
        """Forward pass of the self-attention."""
        # (B, H*W, proj_dim) -> (B*num_heads, H*W, proj_dim//num_heads).
        query, key, value = self.to_qkv(features, context)

        # compute self-attention. Output has same shape as q, k, v
        features = self.self_attn(query, key, value)

        # (B*num_heads, H*W, proj_dim//num_heads) -> (B, H*W, proj_dim)
        features = self._batch2heads(features)

        # linear projection + dropout
        out = self.to_out(features)  # (B, H*W, query_dim)
        if self.dropout is not None:
            out = self.dropout(out)

        return out


class SelfAttentionBlock(nn.Module):
    def __init__(
        self,
        how: str,
        query_dim: int,
        name: str = "exact",
        cross_attention_dim: int = None,
        num_heads: int = 8,
        head_dim: int = 64,
        dropout: float = 0.0,
        bias: bool = False,
        slice_size: int = 4,
        **kwargs,
    ) -> None:
        """Singular self-attention block.

        These can be stacked to form a tranformer block.

        NOTE: Can be used as a cross-attention block if `cross_attention_dim` is given.

        Input Shape: (B, H'*W', query_dim).
        Output Shape: (B, H'*W', query_dim).

        Parameters
        ----------
            name : str
                Name of the attention method. One of ("exact", "linformer").
            how : str, default="basic"
                How to compute the self-attention matrix.
                One of ("basic", "flash", "slice", "memeff", "slice-memeff").
                "basic": the normal O(N^2) self attention.
                "flash": the flash attention (by xformers library),
                "slice": batch sliced attention operation to save mem.
                "memeff": xformers.memory_efficient_attention.
                "slice-memeff": Conmbine slicing and memory_efficient_attention.
            query_dim : int
                The number of channels in the query. Typically: num_heads*head_dim
            cross_attention_dim : int, optional
                Number of channels in the context tensor. Cross attention combines
                asymmetrically two separate embeddings (context and input embeddings).
                E.g. passage from transformer encoder to transformer decoder. If this is
                set to None, no cross attention is applied.
            num_heads : int, default=8
                Number of heads for multi-head attention.
            head_dim : int, default=64
                Number of output channels per head.
            dropout : float, default=0.0
                Dropout probability.
            bias : bool, default=False
                Flag to set bias for Q, K and V.
            slice_size : int, default=4
                Slice size for sliced self-attention. This is used only if
                `self_attention = "slice"`.
        """
        super().__init__()

        self.norm = nn.LayerNorm(query_dim)
        self.att = SelfAttention(
            name=name,
            how=how,
            query_dim=query_dim,
            cross_attention_dim=cross_attention_dim,
            head_dim=head_dim,
            num_heads=num_heads,
            dropout=dropout,
            bias=bias,
            slice_size=slice_size,
            **kwargs,
        )

    def forward(self, x: torch.Tensor, context: torch.Tensor = None) -> torch.Tensor:
        """Forward pass of one transformer block."""
        residual = x
        x = self.norm(x)
        x = self.att(x, context) + residual  # residual post-layernorm

        return x



import warnings
from abc import ABC, abstractmethod
from collections import OrderedDict
from itertools import chain
from pathlib import Path
from typing import Callable, Dict, List, Tuple, Union

import numpy as np
import torch
import torch.nn as nn
import yaml
from torch.utils.data import DataLoader
from tqdm import tqdm

from ..utils import FileHandler, tensor_to_ndarray
from .folder_dataset_infer import FolderDatasetInfer
from .hdf5_dataset_infer import HDF5DatasetInfer
from .post_processor import PostProcessor
from .predictor import Predictor


class BaseInferer(ABC):
    def __init__(
        self,
        model: nn.Module,
        input_path: Union[Path, str],
        out_activations: Dict[str, str],
        out_boundary_weights: Dict[str, bool],
        patch_size: Tuple[int, int],
        instance_postproc: str,
        padding: int = None,
        batch_size: int = 8,
        normalization: str = None,
        device: str = "cuda",
        n_devices: int = 1,
        save_intermediate: bool = False,
        save_dir: Union[Path, str] = None,
        save_format: str = ".mat",
        checkpoint_path: Union[Path, str] = None,
        n_images: int = None,
        type_post_proc: Callable = None,
        sem_post_proc: Callable = None,
        **kwargs,
    ) -> None:
        """Inference for an image folder.

        Parameters
        ----------
            model : nn.Module
                A segmentation model.
            input_path : Path | str
                Path to a folder of images or to hdf5 db.
            out_activations : Dict[str, str]
                Dictionary of head names mapped to a string value that specifies the
                activation applied at the head. E.g. {"type": "tanh", "cellpose": None}
                Allowed values: "softmax", "sigmoid", "tanh", None.
            out_boundary_weights : Dict[str, bool]
                Dictionary of head names mapped to a boolean value. If the value is
                True, after a prediction, a weight matrix is applied that assigns bigger
                weight on pixels in the center and less weight to pixels on the tile
                boundaries. helps dealing with prediction artefacts on the boundaries.
                E.g. {"type": False, "cellpose": True}
            patch_size : Tuple[int, int]:
                The size of the input patches that are fed to the segmentation model.
            instance_postproc : str
                The post-processing method for the instance segmentation mask. One of:
                "cellpose", "omnipose", "stardist", "hovernet", "dcan", "drfns", "dran"
            padding : int, optional
                The amount of reflection padding for the input images.
            batch_size : int, default=8
                Number of images loaded from the folder at every batch.
            normalization : str, optional
                Apply img normalization at forward pass (Same as during training).
                One of: "dataset", "minmax", "norm", "percentile", None.
            device : str, default="cuda"
                The device of the input and model. One of: "cuda", "cpu"
            n_devices : int, default=1
                Number of devices (cpus/gpus) used for inference.
                The model will be copied into these devices.
            save_dir : bool, optional
                Path to save directory. If None, no masks will be saved to disk as .mat
                or .json files. Instead the masks will be saved in `self.out_masks`.
            save_intermediate : bool, default=False
                If True, intermediate soft masks will be saved into `soft_masks` var.
            save_format : str, default=".mat"
                The file format for the saved output masks. One of (".mat", ".json").
                The ".json" option will save masks into geojson format.
            checkpoint_path : Path | str, optional
                Path to the model weight checkpoints.
            n_images : int, optional
                First n-number of images used from the `input_path`.
            type_post_proc : Callable, optional
                A post-processing function for the type maps. If not None, overrides
                the default.
            sem_post_proc : Callable, optional
                A post-processing function for the semantc seg maps. If not None,
                overrides the default.
            **kwargs:
                Arbitrary keyword arguments expecially for post-processing and saving.
        """
        # basic inits
        self.model = model
        self.out_heads = self._get_out_info()  # the names and num channels of out heads
        self.batch_size = batch_size
        self.patch_size = patch_size
        self.padding = padding
        self.out_activations = out_activations
        self.out_boundary_weights = out_boundary_weights
        self.head_kwargs = self._check_and_set_head_args()
        self.kwargs = kwargs

        self.save_dir = Path(save_dir) if save_dir is not None else None
        self.save_intermediate = save_intermediate
        self.save_format = save_format

        # dataset & dataloader
        self.path = Path(input_path)
        if self.path.is_dir():
            ds = FolderDatasetInfer(self.path, n_images=n_images)
            if self.save_dir is None and len(ds.fnames) > 40:
                warnings.warn(
                    "`save_dir` is None. Thus, the outputs are be saved in `out_masks` "
                    "class attribute. If the input folder contains many images, running"
                    " inference will likely flood the memory depending on the size and "
                    "number of the images. Consider saving outputs to disk by providing"
                    " `save_dir` argument."
                )
        elif self.path.is_file() and self.path.suffix in (".h5", ".hdf5"):
            ds = HDF5DatasetInfer(self.path, n_images=n_images)
        else:
            raise ValueError(
                f"Given `input_path`: {input_path} is neither an image folder or a h5 "
                "database. Allowed suffices for h5 database are ('.h5', '.hdf5')"
            )

        self.dataloader = DataLoader(
            ds, batch_size=batch_size, shuffle=False, pin_memory=True
        )

        # Set post processor
        self.postprocessor = PostProcessor(
            instance_postproc,
            inst_key=self.model.inst_key,
            aux_key=self.model.aux_key,
            type_post_proc=type_post_proc,
            sem_post_proc=sem_post_proc,
            **kwargs,
        )

        # load weights and set devices
        if checkpoint_path is not None:
            ckpt = torch.load(
                checkpoint_path, map_location=lambda storage, loc: storage
            )

            try:
                self.model.load_state_dict(ckpt["state_dict"], strict=True)
            except RuntimeError:
                new_ckpt = self._strip_state_dict(ckpt)
                self.model.load_state_dict(new_ckpt["state_dict"], strict=True)
            except BaseException as e:
                print(e)

        assert device in ("cuda", "cpu")
        if device == "cpu":
            self.device = torch.device("cpu")
        if torch.cuda.is_available() and device == "cuda":
            self.device = torch.device("cuda")

            if torch.cuda.device_count() > 1 and n_devices > 1:
                self.model = nn.DataParallel(self.model, device_ids=range(n_devices))

        self.model.to(self.device)
        self.model.eval()

        # Helper class to perform forward + extra processing
        self.predictor = Predictor(
            model=self.model,
            patch_size=patch_size,
            normalization=normalization,
            device=self.device,
        )

    @classmethod
    def from_yaml(cls, model: nn.Module, yaml_path: str):
        """Initialize the inferer from a yaml-file.

        Parameters
        ----------
            model : nn.Module
                Initialized segmentation model.
            yaml_path : str
                Path to the yaml file containing rest of the params
        """
        with open(yaml_path, "r") as stream:
            kwargs = yaml.full_load(stream)

        return cls(model, **kwargs)

    @abstractmethod
    def _infer_batch(self):
        raise NotImplementedError

    def infer(self) -> None:
        """Run inference and post-processing for the images.

        NOTE:
        - Saves outputs in `self.out_masks` or to disk (.mat/.json) files.
        - If `save_intermediate` is set to True, also intermiediate model outputs are
            saved to `self.soft_masks`
        - `self.out_masks` and `self.soft_masks` are nested dicts: E.g.
                {"sample1": {"inst": [H, W], "type": [H, W], "sem": [H, W]}}
        - If masks are saved to geojson .json files, more key word arguments
            need to be given at class initialization. Namely: `geo_format`,
            `classes_type`, `classes_sem`, `offsets`. See more in the
            `FileHandler.save_masks` docs.
        """
        self.soft_masks = {}
        self.out_masks = {}
        self.elapsed = []
        self.rate = []
        with tqdm(self.dataloader, unit="batch") as loader:
            with torch.no_grad():
                for data in loader:
                    names = data["file"]
                    loader.set_description("Running inference")
                    loader.set_postfix_str("Forward pass")
                    soft_masks = self._infer_batch(data["im"])
                    loader.set_postfix_str("post-processing")
                    soft_masks = self._prepare_mask_list(names, soft_masks)

                    # use multi-threading if batch size more than 1.
                    if self.batch_size > 1:
                        seg_results = self.postprocessor.run_parallel(soft_masks)
                    else:
                        seg_results = [
                            self.postprocessor.post_proc_pipeline(soft_masks[0])
                        ]

                    self.elapsed.append(loader.format_dict["elapsed"])
                    self.rate.append(loader.format_dict["rate"])

                    if self.save_intermediate:
                        for n, m in zip(names, soft_masks):
                            self.soft_masks[n] = m

                    if self.save_dir is None:
                        for n, m in zip(names, seg_results):
                            self.out_masks[n] = m
                    else:
                        loader.set_postfix_str("Saving results to disk")
                        if self.batch_size > 1:
                            fnames = [Path(self.save_dir) / n for n in names]
                            FileHandler.save_masks_parallel(
                                maps=seg_results,
                                fnames=fnames,
                                **{**self.kwargs, "format": self.save_format},
                            )
                        else:
                            for n, m in zip(names, seg_results):
                                fname = Path(self.save_dir) / n
                                FileHandler.save_masks(
                                    fname=fname,
                                    maps=m,
                                    **{**self.kwargs, "format": self.save_format},
                                )

    def _strip_state_dict(self, ckpt: Dict) -> OrderedDict:
        """Strip te first 'model.' (generated by lightning) from the state dict keys."""
        state_dict = OrderedDict()
        for k, w in ckpt["state_dict"].items():
            if "num_batches_track" not in k:
                new_key = k.strip("model")[1:]
                state_dict[new_key] = w
        ckpt["state_dict"] = state_dict

        return ckpt

    def _check_and_set_head_args(self) -> None:
        """Check the model has matching head names with the head args and set them."""
        heads = [head_name for head_name, _ in self.out_heads]
        if self.out_boundary_weights.keys() != self.out_activations.keys():
            raise ValueError(
                "Mismatching head names in `out_boundary_weights` & `out_activations`",
            )

        pred_kwargs = {}
        for head_name, val in self.out_activations.items():
            if head_name in heads:
                pred_kwargs[head_name] = {"act": val}
                pred_kwargs[head_name]["apply_weights"] = self.out_boundary_weights[
                    head_name
                ]
            else:
                raise ValueError(
                    f"Mismatching head name. The model contains heads: {heads} ",
                    f"Got: {head_name}",
                )

        return pred_kwargs

    def _prepare_mask_list(
        self, fnames: List[str], soft_masks: torch.Tensor
    ) -> List[Dict[str, np.ndarray]]:
        """Convert the model outputs into arguments for parallel post-processing."""
        softies = {}
        for k, soft_mask in soft_masks.items():
            softies[k] = dict(zip(fnames, tensor_to_ndarray(soft_mask, "BCHW", "BCHW")))

        args = []
        for fn in fnames:
            map_dict = {}
            for mn, d in softies.items():
                map_dict[mn] = d[fn]
            args.append(map_dict)

        return args

    def _get_out_info(self) -> Tuple[Tuple[str, int]]:
        """Get the output names and number of out channels."""
        return tuple(
            chain.from_iterable(
                list(self.model.heads[k].items()) for k in self.model.heads.keys()
            )
        )



"""
Imported from cellpose repo. Refacotred for readability.

https://github.com/MouseLand/cellpose/blob/master/cellpose/dynamics.py

Copyright © 2020 Howard Hughes Medical Institute

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer. Redistributions in
binary form must reproduce the above copyright notice, this list of
conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution. Neither the name of HHMI
nor the names of its contributors may be used to endorse or promote
products derived from this software without specific prior written
permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
“AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
from typing import List, Tuple

import numpy as np
from scipy.ndimage import maximum_filter1d
from skimage.filters import apply_hysteresis_threshold

from cellseg_models_pytorch.utils import binarize

from .integrator import follow_flows
from .utils import fill_holes_and_remove_small_masks, gen_flows

__all__ = ["post_proc_cellpose", "get_masks_cellpose"]


def get_seeds(
    p: np.ndarray, rpad: int = 20, dims: int = 2
) -> Tuple[Tuple[np.ndarray, np.ndarray], np.ndarray, List[np.ndarray]]:
    """Get seed points to cells from the 2D-histogram of the flows.

    Return also the histogram and flows.

    Parameters
    ----------
        p : np.ndarray
            The computed flows. Shape (2, H, W). Dtype: float32.
        rpad : int, default=20
            The amount of padding when computing seeds.
        dims : int, default=2
            The number of dimensions in the input data.

    Returns
    -------
        Tuple[Tuple[np.ndarray, np.ndarray], np.ndarray, List[np.ndarray]]:
            seeds : Tuple[np.ndarray, np.ndarray]
                The y- and x- seeds sorted by descending bincount.
                np.ndarray.shape: (n_cells, ). Dtype: int32.
            h : np.ndarray
                The 2D histogram of the flows p. Shape (H+pad, W+pad).
            pflows : List[np.ndarray]
                The flattend y and x flows. Array shape: (H*W).
    """
    shape = p.shape[1:]
    dims = len(p)

    pflows = []
    edges = []

    for i in range(dims):
        pflows.append(np.int32(p[i]).flatten())
        edges.append(np.arange(-0.5 - rpad, shape[i] + 0.5 + rpad, 1))  # bins

    h, _ = np.histogramdd(
        tuple(pflows), bins=edges
    )  # (cell seeds) as histogram bins (H, W)
    hmax = h.copy()
    for i in range(dims):
        hmax = maximum_filter1d(hmax, 5, axis=i)  # enlarge the seeds

    # non-zero pixel yx-coords of the bins as seeds
    seeds: Tuple[np.ndarray, np.ndarray] = np.nonzero(
        np.logical_and(h - hmax > -1e-6, h > 10)
    )

    Nmax: np.ndarray = h[seeds]  # bincounts, len == n cells
    isort: np.ndarray = np.argsort(Nmax)[::-1]  # sorted bincount indices, descending
    seeds = [s[isort] for s in seeds]  # sort yx-coords by bincount, descending

    return seeds, h, pflows


def expand_seed_pixels(
    seeds: Tuple[np.ndarray, np.ndarray], h: np.ndarray, dims: int = 2
) -> Tuple[Tuple[np.ndarray, np.ndarray]]:
    """Expand the seed pixels to 3x3 neihgborhood of pixels for every seed.

    Parameters
    ----------
        seeds : Tuple[np.ndarray, np.ndarray]
            The y- and x- seeds sorted by descending bincount.
            np.ndarray.shape: (n_cells, ). Dtype: int32.
        h : np.ndarray
            The 2D histogram of the flows. Shape (H+pad, W+pad).
        dims : int, default=2
            The number of dimensions in the input data.

    Returns
    -------
        Tuple[Tuple[np.ndarray, np.ndarray]]:
            The expanded pixel neighborhood coords for all the seeds
    """
    pix: List[Tuple[np.ndarray, np.ndarray]] = list(np.array(seeds).T)
    shape: Tuple[int, int] = h.shape
    expand: Tuple[np.ndarray, np.ndarray] = np.nonzero(np.ones((3, 3)))

    # loop the seed coords
    for k in range(len(pix)):

        # convert tuple to list
        pix[k] = list(pix[k])

        newpix = []
        iin = []
        for i, e in enumerate(expand):
            # centered 1d window around pixel x- or y-coords
            # `pix[k]` can be a  varying length
            epix: np.ndarray = (
                e[:, None] + np.expand_dims(pix[k][i], 0) - 1
            )  # (9, n_pixels)
            epix: np.ndarray = epix.flatten()

            # make sure window is inside the image
            iin.append(np.logical_and(epix >= 0, epix < shape[i]))
            newpix.append(epix)

        # filter pixls whose y and x coords inside img
        iin = np.all(tuple(iin), axis=0)
        newpix = [pi[iin] for pi in newpix]

        # include all pixels with more than 2 final pixels p of the newpix window
        newpix = tuple(newpix)
        igood = h[newpix] > 2
        for i in range(dims):
            pix[k][i] = newpix[i][igood]

        # convert back to tuple
        pix[k] = tuple(pix[k])

    return pix


def get_masks_cellpose(p: np.ndarray, rpad: int = 20) -> np.ndarray:
    """Create masks using pixel convergence after running dynamics.

    Makes a histogram of final pixel locations p, initializes masks
    at peaks of histogram and extends the masks from the peaks so that
    they include all pixels with more than 2 final pixels p. Discards
    masks with flow errors greater than the threshold.

    Parameters
    ----------
        p : np.ndarray
            Final locations of each pixel after dynamics. Shape (2, H, W).
        rpad : int, default=20
            Histogram edge padding.

    Returns
    -------
        np.ndarray:
            Instance labelled mask. Shape (H, W).
    """
    shape0 = p.shape[1:]
    dims = len(p)

    seeds, h, pflows = get_seeds(p, rpad, dims)
    pix = expand_seed_pixels(seeds, h, dims)

    M = np.zeros(h.shape, np.int32)
    for k in range(len(pix)):
        M[pix[k]] = 1 + k

    for i in range(dims):
        pflows[i] = pflows[i] + rpad

    # remove big masks
    M0 = M[tuple(pflows)]
    _, counts = np.unique(M0, return_counts=True)
    big = float(np.prod(shape0))
    for i in np.nonzero(counts > big)[0]:
        M0[M0 == i] = 0

    _, M0 = np.unique(M0, return_inverse=True)
    M0 = np.reshape(M0, shape0)

    return M0


def post_proc_cellpose(
    inst_map: np.ndarray,
    flow_map: np.ndarray,
    dist_map: np.ndarray = None,
    return_flows: bool = False,
    min_size: int = 30,
    interp: bool = True,
    use_gpu: bool = True,
    **kwargs
) -> np.ndarray:
    """Run the cellpose post-processing pipeline.

    https://www.nature.com/articles/s41592-020-01018-x

    Parameters
    ----------
        inst_map : np.ndarray
            Instance labelled or binary mask. Shape (H, W).
        flow_map : np.ndarray
            Y- and x-flows. Shape: (2, H, W)
        dist_map : np.ndarray, default=None
            Regressed distance transform. Shape: (H, W).
        return_flows : bool, default=False
            If True, returns the HSV converted flows. They are just not
            needed for anything relevant.
        min_size : int
            The minimum size for the objects that will not be removed.
        interp : bool, default=True
            Use bilinear interpolation when integrating the flow dynamics.
        use_gpu : bool, default=True
            Use gpu accelerated bilinear interpolation. If `interp` == False, this is
            ignored.

    Returns
    -------
        np.ndarray:
            The instance labelled segmentation mask. Shape (H, W)
    """
    #  convert channels to CHW
    if dist_map is not None:
        binary_mask = apply_hysteresis_threshold(dist_map, 0.5, 0.5)
    else:
        binary_mask = binarize(inst_map).astype(bool)

    dP = flow_map * binary_mask  # /5.
    # dP = normalize_field(dP)

    pixel_loc, _ = follow_flows(
        dP,
        niter=200,
        mask=binary_mask,
        suppress_euler=False,
        interp=interp,
        use_gpu=use_gpu,
    )

    mask = get_masks_cellpose(pixel_loc)
    inst_map = fill_holes_and_remove_small_masks(mask, min_size=min_size).astype("i4")

    if return_flows:
        hsv_flows = gen_flows(dP)
        return inst_map, hsv_flows

    return inst_map



import torch

from .attention_ops import compute_mha
from .base_attention import BaseSelfAttention

__all__ = ["ExactSelfAttention"]


class ExactSelfAttention(BaseSelfAttention):
    def __init__(
        self,
        head_dim: int,
        num_heads: int,
        how: str = "basic",
        slice_size: int = None,
        **kwargs,
    ) -> None:
        """Compute exact attention.

        Four variants:
        - basic: self-attention implementation with torch.matmul O(N^2)
        - slice-attention: Computes the attention matrix in slices to save mem.
        - memeff: `xformers.ops.memory_efficient_attention` from xformers package.
        - slice-memeff-attention: Comnbines slice-attention and memeff

        Parameters
        ----------
            head_dim : int
                Out dim per attention head.
            num_heads : int
                Number of heads.
            how : str, default="basic"
                How to compute the self-attention matrix.
                One of ("basic", "flash", "slice", "memeff", "slice-memeff").
                "basic": the normal O(N^2) self attention.
                "flash": the flash attention (by xformers library),
                "slice": batch sliced attention operation to save mem.
                "memeff": xformers.memory_efficient_attention.
                "slice-memeff": Conmbine slicing and memory_efficient_attention.
            slice_size, int, optional
                The size of the slice. Used only if `how in ('slice', 'slice_memeff)`.

        Raises
        ------
            - ValueError:
                - If illegal self attention (`how`) method is given.
                - If `how` is set to `slice` while `num_heads` | `slice_size`
                    args are not given proper integer values.
                - If `how` is set to `memeff` or `slice_memeff` but cuda is not
                    available.
            - ModuleNotFoundError:
                - If `self_attention` is set to `memeff` and `xformers` package is not
                installed
        """
        super().__init__(
            head_dim=head_dim,
            num_heads=num_heads,
            how=how,
            slice_size=slice_size,
        )

    def forward(
        self, query: torch.Tensor, key: torch.Tensor, value: torch.Tensor
    ) -> torch.Tensor:
        """Compute exact self-attention.

        I.e softmax(Q @ K'/sqrt(head_dim)) @ V

        Parameters
        ----------
            query : torch.Tensor
                Query tensor. Shape: (B*num_heads, H*W, proj_dim//num_heads).
            key : torch.Tensor
                Key tensor. Shape: (B*num_heads, H*W, proj_dim//num_heads).
            value : torch.Tensor
                Value tensor. Shape: (B*num_heads, H*W, proj_dim//num_heads).

        Returns
        -------
            torch.Tensor:
                The self-attention matrix. Same shape as inputs.
        """
        attn = compute_mha(
            query,
            key,
            value,
            self.how,
            slice_size=self.slice_size,  # used only for slice-att
            num_heads=self.num_heads,  # used only for slice-att
            proj_channels=self.proj_channels,  # used only for slice-att
        )

        return attn

    def __repr__(self) -> str:
        """Add extra info to print."""
        s = "ExactSelfAttention(how='{how}', head_dim={head_dim}, num_heads={num_heads})"  # noqa: E501
        s = s.format(**self.__dict__)
        return s



import torch
import torch.nn as nn

from .act import ACT_LOOKUP
from .conv import CONV_LOOKUP
from .norm import NORM_LOOKUP
from .self_attention import SELFATT_LOOKUP
from .upsample import UP_LOOKUP

__all__ = ["Activation", "Norm", "Up", "Conv", "Identity", "MultiHeadSelfAttention"]


class Identity(nn.Module):
    def __init__(self, in_channels: int = None, *args, **kwargs):
        """Identity operator that is argument-insensitive.

        The forward method can take in multiple arguments returning only the
        first one.
        """
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = in_channels

    def forward(self, input: torch.Tensor, *args, **kwargs) -> torch.Tensor:
        """Forward pass of idenStity operator."""
        return input


class Activation(nn.Module):
    def __init__(self, name: str, **kwargs) -> None:
        """Activation wrapper class.

        Parameters:
        -----------
            name : str
                Name of the activation method.

        Raises
        ------
            ValueError: if the activation method name is illegal.
        """
        super().__init__()

        allowed = list(ACT_LOOKUP.keys()) + [None]
        if name not in allowed:
            raise ValueError(
                f"Illegal activation method given. Allowed: {allowed}. Got: '{name}'"
            )

        if name is not None:
            try:
                self.act = ACT_LOOKUP[name](**kwargs, inplace=True)
            except Exception:
                try:
                    self.act = ACT_LOOKUP[name](**kwargs)
                except Exception as e:
                    raise Exception(
                        "Encountered an error when trying to init activation function: "
                        f"Activation(name='{name}'): {e.__class__.__name__}: {e}"
                    )
        else:
            self.act = Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of the activation function."""
        return self.act(x)


class Norm(nn.Module):
    def __init__(self, name: str, **kwargs) -> None:
        """Normalize wrapper class.

        Parameters:
        -----------
            name : str
                Name of the normalization method.

        Raises
        ------
            ValueError: if the normalization method name is illegal.
        """
        super().__init__()

        allowed = list(NORM_LOOKUP.keys()) + [None]
        if name not in allowed:
            raise ValueError(
                f"Illegal normalization method given. Allowed: {allowed}. Got: '{name}'"
            )

        if name is not None:
            try:
                self.norm = NORM_LOOKUP[name](**kwargs)
            except Exception as e:
                raise Exception(
                    "Encountered an error when trying to init normalization function: "
                    f"Norm(name='{name}'): {e.__class__.__name__}: {e}"
                )
        else:
            self.norm = Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass for the norm function."""
        return self.norm(x)


class Up(nn.Module):
    def __init__(self, name: str, scale_factor: int = 2, **kwargs) -> None:
        """Upsample wrapper class.

        Parameters:
        -----------
            name : str
                Name of the upsampling method.
            scale_factor : int, default=2
                Upsampling scale factor. scale_factor*(H, W)

        Raises
        ------
            ValueError: if the upsampling method name is illegal.
        """
        super().__init__()

        allowed = list(UP_LOOKUP.keys())
        if name not in allowed:
            raise ValueError(
                f"Illegal upsampling method given. Allowed: {allowed}. Got: '{name}'"
            )

        if name in ("bilinear", "bicubic"):
            kwargs["mode"] = name
            kwargs["align_corners"] = True

        kwargs["scale_factor"] = scale_factor

        try:
            self.up = UP_LOOKUP[name](**kwargs)
        except Exception as e:
            raise Exception(
                "Encountered an error when trying to init upsampling function: "
                f"Up(name='{name}'): {e.__class__.__name__}: {e}"
            )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass for the upsampling function."""
        return self.up(x)


class Conv(nn.Module):
    def __init__(self, name: str, **kwargs) -> None:
        """Convolution wrapper class.

        Parameters:
        -----------
            name : str
                Name of the convolution method.

        Raises
        ------
            ValueError: if the convolution method name is illegal.
        """
        super().__init__()

        allowed = list(CONV_LOOKUP.keys())
        if name not in allowed:
            raise ValueError(
                f"Illegal convolution method given. Allowed: {allowed}. Got: '{name}'"
            )

        try:
            self.conv = CONV_LOOKUP[name](**kwargs)
        except Exception as e:
            raise Exception(
                "Encountered an error when trying to init convolution function: "
                f"Conv(name='{name}'): {e.__class__.__name__}: {e}"
            )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass for the convolution function."""
        return self.conv(x)


class MultiHeadSelfAttention(nn.Module):
    def __init__(self, name: str, **kwargs) -> None:
        """Multi-head self-attention wrapper class.

        Parameters:
        -----------
            name : str
                Name of the mhsa method.

        Raises
        ------
            ValueError: if the mhsa method name is illegal.
        """
        super().__init__()

        allowed = list(SELFATT_LOOKUP.keys())
        if name not in allowed:
            raise ValueError(
                "Illegal multi-head attention method given. "
                f"Allowed: {allowed}. Got: '{name}'"
            )

        try:
            self.att = SELFATT_LOOKUP[name](**kwargs)
        except Exception as e:
            raise Exception(
                "Encountered an error when trying to init convolution function: "
                f"MultiHeadSelfAttention(name='{name}'): {e.__class__.__name__}: {e}"
            )

    def forward(
        self,
        query: torch.Tensor,
        key: torch.Tensor,
        value: torch.Tensor,
        **kwargs,
    ) -> torch.Tensor:
        """Forward pass for the convolution function."""
        return self.att(query, key, value, **kwargs)



from torch.nn import BatchNorm2d, InstanceNorm2d, LayerNorm, SyncBatchNorm

from .bcn import BCNorm
from .gn import GroupNorm
from .ln import LayerNorm2d

NORM_LOOKUP = {
    "bn": BatchNorm2d,
    "ln2d": LayerNorm2d,
    "ln": LayerNorm,
    "bcn": BCNorm,
    "gn": GroupNorm,
    "in": InstanceNorm2d,
}

__all__ = [
    "NORM_LOOKUP",
    "BCNorm",
    "GroupNorm",
    "BatchNorm2d",
    "InstanceNorm2d",
    "SyncBatchNorm",
    "LayerNorm2d",
    "LayerNorm",
]



import torch
import torch.nn as nn

__all__ = ["StarReLU"]


class StarReLU(nn.Module):
    def __init__(
        self,
        scale_value: float = 1.0,
        bias_value: float = 0.0,
        scale_learnable: bool = True,
        bias_learnable: bool = True,
        inplace: bool = False,
        **kwargs
    ) -> None:
        """Apply StarReLU activation.

        Adapted from:
        https://github.com/sail-sg/metaformer/blob/main/metaformer_baselines.py

        See MetaFormer: https://arxiv.org/abs/2210.13452

        StarReLU: s * relu(x) ** 2 + b

        Parameters
        ----------
            scale_value : float, default=1.0
                Learnable scaling factor for relu activation.
            bias_value : float, default=0.0
                Learnable bias term for relu activation.
            scale_learnable : bool, default=True
                Flag, whether to keep the scale factor learnable.
            bias_learnable : bool, default=True
                Flag, whether to keep the bias term learnable.
            inplace : bool, default=False
                Flag whether to apply inplace-relu.
        """
        super().__init__()
        self.inplace = inplace
        self.relu = nn.ReLU(inplace=inplace)
        self.scale = nn.Parameter(
            scale_value * torch.ones(1), requires_grad=scale_learnable
        )
        self.bias = nn.Parameter(
            bias_value * torch.ones(1), requires_grad=bias_learnable
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of the StarReLU."""
        return self.scale * self.relu(x) ** 2 + self.bias



import torch

try:
    from xformers.ops import memory_efficient_attention

    _has_xformers = True
except ModuleNotFoundError:
    _has_xformers = False


__all__ = ["multihead_attention", "slice_mha", "mha", "compute_mha"]


def multihead_attention(
    query: torch.Tensor,
    key: torch.Tensor,
    value: torch.Tensor,
    scale: float = None,
    **kwargs,
) -> torch.Tensor:
    """Compute exact self attention with torch. Complexity: O(N**2).

    Parameters
    ----------
        query : torch.Tensor
            Query tensor. Shape: (B*num_heads, H*W, proj_dim//num_heads).
        key : torch.Tensor
            Key tensor. Shape: (B*num_heads, H*W, proj_dim//num_heads).
        value : torch.Tensor
            Value tensor. Shape: (B*num_heads, H*W, proj_dim//num_heads).
        scale : float, optional
            Scaling factor for Q @ K'. If None, query.shape[-1]**-0.5 will
            be used

    Returns
    -------
        torch.Tensor:
            The self-attention matrix. Same shape as inputs.
    """
    if scale is None:
        scale = query.shape[-1] ** -0.5

    scores = torch.matmul(query, key.transpose(-1, -2)) * scale
    probs = scores.softmax(dim=-1)

    # compute attention output
    return torch.matmul(probs, value)


def mha(
    query: torch.Tensor,
    key: torch.Tensor,
    value: torch.Tensor,
    att_type: str = "basic",
    **kwargs,
) -> torch.Tensor:
    """Compute exact self-attention.

    I.e softmax(Q @ K'/sqrt(head_dim)) @ V

    Parameters
    ----------
        query : torch.Tensor
            Query tensor. Shape: (B*num_heads, H*W, proj_dim//num_heads).
        key : torch.Tensor
            Key tensor. Shape: (B*num_heads, H*W, proj_dim//num_heads).
        value : torch.Tensor
            Value tensor. Shape: (B*num_heads, H*W, proj_dim//num_heads).
        att_type : str, default="basic"
            The type of the self-attention computation.
            One of: ("basic", "flash", "memeff").
        **kwargs:
            Extra key-word arguments for the mha computation.

    Returns
    -------
        torch.Tensor:
            The self-attention matrix. Same shape as inputs.
    """
    if att_type == "memeff":
        if _has_xformers:
            if all([query.is_cuda, key.is_cuda, value.is_cuda]):
                attn = memory_efficient_attention(query, key, value)
            else:
                raise RuntimeError(
                    "`xformers.ops.memory_efficient_attention` is only implemented "
                    "for cuda. Make sure your inputs & model devices are set to cuda."
                )
        else:
            raise ModuleNotFoundError(
                "Trying to use `memory_efficient_attention`. The method requires the "
                "xformers package. See how to install xformers: "
                "https://github.com/facebookresearch/xformers"
            )
    elif att_type == "flash":
        raise NotImplementedError
    elif att_type == "basic":
        attn = multihead_attention(query, key, value, **kwargs)
    else:
        raise ValueError(
            f"Unknown `att_type` given. Got: {att_type}. "
            f"Allowed: {('memeff', 'flash', 'basic')}"
        )

    return attn


def slice_mha(
    query: torch.Tensor,
    key: torch.Tensor,
    value: torch.Tensor,
    proj_channels: int,
    num_heads: int,
    slice_size: int = 4,
    att_type: str = "basic",
    **kwargs,
) -> torch.Tensor:
    """Compute exact attention in slices to save memory.

    NOTE: adapted from hugginface diffusers package. Their implementation
    just dont handle the case where B // slize_size doesn't divide evenly.
    This would end up in zero-matrices at the final batch dimensions of
    the batched attention matrix.

    NOTE: The input is sliced in the batch dimension.

    Parameters
    ----------
        query : torch.Tensor
            Query tensor. Shape: (B*num_heads, H*W, proj_dim//num_heads).
        key : torch.Tensor
            Key tensor. Shape: (B*num_heads, H*W, proj_dim//num_heads).
        value : torch.Tensor
            Value tensor. Shape: (B*num_heads, H*W, proj_dim//num_heads).
        proj_channels : int
            Number of out channels in the token projections.
        num_heads : int
            Number of heads in the mha.
        slice_size : int, default=4
            The size of the batch dim slice.
        att_type : str, default="basic"
            The type of the self-attention computation.
            One of: ("memeff", "slice-memeff").

    Returns
    -------
        torch.Tensor:
            The self-attention matrix. Same shape as inputs.
    """
    allowed = ("slice", "slice-memeff")
    if att_type not in allowed:
        raise ValueError(
            f"Illegal slice-attention given. Got: {att_type}. Allowed: {allowed}."
        )

    # parse the attention type arg
    a = att_type.split("-")
    if len(a) == 1:
        att_type = "basic"
    else:
        att_type = "memeff"

    B, seq_len = query.shape[:2]
    out = torch.zeros(
        (B, seq_len, proj_channels // num_heads),
        device=query.device,
        dtype=query.dtype,
    )

    # get the modulo if B/slice_size is not evenly divisible.
    n_slices, mod = divmod(out.shape[0], slice_size)
    if mod != 0:
        n_slices += 1

    it = list(range(n_slices))
    for i in it:
        start = i * slice_size
        end = (i + 1) * slice_size

        if i == it[-1]:
            end = start + mod

        attn_slice = mha(
            query[start:end], key[start:end], value[start:end], att_type=att_type
        )

        out[start:end] = attn_slice
        del attn_slice
        torch.cuda.empty_cache()

    return out


def compute_mha(
    query: torch.Tensor,
    key: torch.Tensor,
    value: torch.Tensor,
    how: str,
    **kwargs,
) -> torch.Tensor:
    """Wrap all the different attention matrix computation types under this.

    Parameters
    ----------
        query : torch.Tensor
            Query tensor. Shape: (B*num_heads, H*W, proj_dim//num_heads).
        key : torch.Tensor
            Key tensor. Shape: (B*num_heads, H*W, proj_dim//num_heads).
        value : torch.Tensor
            Value tensor. Shape: (B*num_heads, H*W, proj_dim//num_heads).
        how : str, default="basic"
            How to compute the self-attention matrix.
            One of ("basic", "flash", "slice", "memeff", "slice-memeff").
            "basic": the normal O(N^2) self attention.
            "flash": the flash attention (by xformers library),
            "slice": batch sliced attention operation to save mem.
            "memeff": xformers.memory_efficient_attention.
            "slice-memeff": Conmbine slicing and memory_efficient_attention.
        **kwargs:
            Extra key-word args for the attention matrix computation.
    """
    allowed = ("basic", "flash", "slice", "memeff", "slice-memeff")
    if how not in allowed:
        raise ValueError(
            f"Illegal exact self attention type given. Got: {how}. "
            f"Allowed: {allowed}."
        )

    if how == "basic":
        attn = mha(query, key, value, att_type="basic", **kwargs)
    elif how == "memeff":
        attn = mha(query, key, value, att_type="memeff", **kwargs)
    elif how == "slice":
        attn = slice_mha(query, key, value, att_type="slice", **kwargs)
    elif how == "slice-memeff":
        attn = slice_mha(query, key, value, att_type="slice-memeff", **kwargs)
    elif how == "flash":
        raise NotImplementedError
    elif how == "slice-flash":
        raise NotImplementedError

    return attn



import pytest
import torch

from cellseg_models_pytorch.modules import MetaFormer


@pytest.mark.parametrize("type", ["pool", "mscan", "mlp", "self-attention"])
def test_metaformer(type):
    in_channels = 3
    head_dim = 64
    num_heads = 8
    query_dim = head_dim * num_heads
    out_channels = 16

    embed_kwargs = {
        "in_channels": 3,
        "kernel_size": 7,
        "stride": 4,
        "pad": 2,
        "head_dim": head_dim,
        "num_heads": num_heads,
    }

    if type == "self-attention":
        mixer_kwargs = {
            "token_mixer": "self-attention",
            "normalization": "ln",
            "residual": True,
            "norm_kwargs": {"normalized_shape": query_dim},
            "mixer_kwargs": {
                "query_dim": query_dim,
                "name": "exact",
                "how": "basic",
                "cross_attention_dim": None,
            },
        }

    elif type == "mlp":
        mixer_kwargs = {
            "token_mixer": "mlp",
            "normalization": "ln",
            "norm_kwargs": {"normalized_shape": query_dim},
            "mixer_kwargs": {
                "in_channels": query_dim,
            },
        }

    elif type in ("pool", "mscan"):
        mixer_kwargs = {
            "token_mixer": type,
            "normalization": "bn",
            "norm_kwargs": {
                "num_features": query_dim,
            },
            "mixer_kwargs": {
                "kernel_size": 3,
                "in_channels": query_dim,
            },
        }

    mlp_kwargs = {
        "in_channels": query_dim,
        "norm_kwargs": {"normalized_shape": query_dim},
    }

    metaformer = MetaFormer(
        in_channels=in_channels,
        out_channels=out_channels,
        embed_kwargs=embed_kwargs,
        mixer_kwargs=mixer_kwargs,
        mlp_kwargs=mlp_kwargs,
        layer_scale=True,
        dropout=0.1,
    )

    x = torch.rand([8, 3, 32, 32])
    dd = metaformer(x)

    assert dd.shape == torch.Size([8, out_channels, 32, 32])



import torch
import torch.nn as nn
import torch.nn.functional as F

from .attention_ops import compute_mha
from .base_attention import BaseSelfAttention

__all__ = ["LinformerAttention"]


class LinformerAttention(BaseSelfAttention):
    def __init__(
        self,
        seq_len: int,
        head_dim: int,
        num_heads: int,
        k: int = None,
        how: str = "basic",
        slice_size: int = None,
        **kwargs
    ) -> None:
        """Linformer attention mechanism.

        Linformer: Self-Attention with Linear Complexity
        - https://arxiv.org/abs/2006.04768v2

        Adapted from xformers library

        NOTE: Weirdly, even when computing linformer attention with xformers
        `memory_efficient_attention`, linformer needs more memory for long sequences
        (due to the linear layers) than computing exact `memory_efficient_attention`.

        Parameters
        ----------
            seq_len : int
                The length of the sequence. (For per-pixel patches H*W).
            head_dim : int
                Out dim per attention head.
            num_heads : int
                Number of heads.
            k : int, optional
                Divisor for key and value matrices to get low-rank attention matrix.
            how : str, default="basic"
                How to compute the self-attention matrix.
                One of ("basic", "flash", "slice", "memeff", "slice_memeff").
                "basic": the normal O(N^2) self attention.
                "flash": the flash attention (by xformers library),
                "slice": batch sliced attention operation to save mem.
                "memeff": xformers.memory_efficient_attention.
                "slice_memeff": Conmbine slicing and memory_efficient_attention.
            slice_size, int, optional
                The size of the slice. Used only if `how in ('slice', 'slice_memeff)`.
        """
        super().__init__(
            head_dim=head_dim,
            num_heads=num_heads,
            how=how,
            slice_size=slice_size,
        )

        if k is None:
            k = seq_len // 4

        self.k = k
        self.E = nn.Linear(seq_len, k, bias=False)
        self.F = nn.Linear(seq_len, k, bias=False)
        self.seq_len = seq_len

    def forward(
        self, query: torch.Tensor, key: torch.Tensor, value: torch.Tensor, **kwargs
    ) -> torch.Tensor:
        """Forward pass of the linformer attention mechanism."""
        padding = 0
        if query.shape[1] < self.seq_len:
            padding = self.seq_len - query.shape[1]
            pad_dims = (0, 0, 0, padding)
            query = F.pad(query, pad_dims)
            key = F.pad(key, pad_dims)
            value = F.pad(value, pad_dims)

        key_proj = self.E(key.transpose(-1, -2)).transpose(-1, -2).contiguous()
        value_proj = self.F(value.transpose(-1, -2)).transpose(-1, -2).contiguous()

        out = compute_mha(
            query,
            key_proj,
            value_proj,
            self.how,
            slice_size=self.slice_size,  # used only for slice-att
            num_heads=self.num_heads,  # used only for slice-att
            proj_channels=self.proj_channels,  # used only for slice-att
        )

        return out[:, :-padding, :] if padding > 0 else out



"""
Adapted `gen_stardist_maps` form the stardist repo.

- https://github.com/stardist/stardist

BSD 3-Clause License

Copyright (c) 2018-2022, Uwe Schmidt, Martin Weigert
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
from typing import Tuple

import numpy as np
import scipy.ndimage as ndi
from numba import njit
from skimage.measure import find_contours, regionprops
from skimage.morphology import dilation, disk

from ...utils import remove_small_objects

__all__ = ["evenly_spaced_points", "gen_radial_distmaps", "gen_stardist_maps"]


@njit
def gen_stardist_maps(inst_map: np.ndarray, n_rays: int) -> np.ndarray:
    """Compute radial distances for each non-zero pixel in a label mask.

    NOTE: Adapted from
    - https://github.com/stardist/stardist/blob/master/stardist/geometry/geom2d.py

    Parameters
    ----------
        inst_map : np.ndarray
            Instance labelled mask. Shape (H, W).
        n_rays : int, default=32
            Number of rays.

    Returns
    -------
        np.ndarray:
            The radial distance maps. Shape (n_rays, H, W). Dtype: float32
    """
    n_rays = int(n_rays)
    dist = np.empty(inst_map.shape + (n_rays,), np.float32)

    st_rays = np.float32((2 * np.pi) / n_rays)
    for i in range(inst_map.shape[0]):
        for j in range(inst_map.shape[1]):
            value = inst_map[i, j]
            if value == 0:
                dist[i, j] = 0
            else:
                for k in range(n_rays):
                    phi = np.float32(k * st_rays)
                    dy = np.cos(phi)
                    dx = np.sin(phi)
                    x, y = np.float32(0), np.float32(0)
                    while True:
                        x += dx
                        y += dy
                        ii = int(round(i + x))
                        jj = int(round(j + y))
                        if (
                            ii < 0
                            or ii >= inst_map.shape[0]
                            or jj < 0
                            or jj >= inst_map.shape[1]
                            or value != inst_map[ii, jj]
                        ):
                            # small correction as we overshoot the boundary
                            t_corr = 1 - 0.5 / max(np.abs(dx), np.abs(dy))
                            x -= t_corr * dx
                            y -= t_corr * dy
                            dst = np.sqrt(x**2 + y**2)
                            dist[i, j, k] = dst
                            break

    return dist.transpose(2, 0, 1)


def evenly_spaced_points(
    binary_map: np.ndarray, n: int
) -> Tuple[np.ndarray, np.ndarray]:
    """Find evenly spaced points on a contour of an object.

    Interpolates evenly spaced points from the arc length of the contour.

    Based on:
    https://stackoverflow.com/questions/27429784/equally-spaced-points-in-a-contour

    Parameters
    ----------
        binary_map : np.ndarray
            A binary mask containing one object/nuclei.
        n : int
            Number of evenly spaced points

    Returns
    -------
        Tuple[np.ndarray, np.ndarray]:
            The y- and x- coordinates of the evenly spaced points. Shapes: (n*).
    """
    n_objs = len(np.unique(binary_map))
    if n_objs > 2:
        raise ValueError(
            f"Expects the input to contain only one object. Got: {n_objs}."
        )

    contours = find_contours(binary_map)[0]
    xc = contours[:, 1]
    yc = contours[:, 0]

    # spacing of x and y points.
    dy = np.diff(yc)
    dx = np.diff(xc)

    # distances between consecutive coordinates
    dS = np.sqrt(dx**2 + dy**2)
    dS = np.append(np.zeros(1), dS)  # include starting point

    # Arc length and perimeter
    d = np.cumsum(dS)
    perim = d[-1]

    # divide the perimeter to evenly spaced values
    ds = perim / n
    dSi = np.arange(0, n) * ds
    dSi[-1] = dSi[-1] - 5e-3

    # interpolate the x and y coordinates
    yi = np.interp(dSi, d, yc)
    xi = np.interp(dSi, d, xc)

    return yi, xi


# TODO: njit this
def gen_radial_distmaps(
    inst_map: np.ndarray, n_rays: int = 32, n_segments: int = 15
) -> np.ndarray:
    """Compute radial distances for each non-zero pixel.

    This is a python (workaround) implementation of the stardist
    radial distance computation algorithm. Results in smoother distance
    maps than the original.

    NOTE: This implementation is still slower than the C implementation
    found in the stardist library. However, a lot faster than the python
    implementation found in the same library. With default params atleast
    x16 faster

    Here, for every object, evenly spaced points are computed from the
    contour of the object and those points are used as line segments.
    The distance for every pixel is computed from the intersection point
    of the rays and line segments rather than shooting rays from each
    pixel.

    Parameters
    ----------
        inst_map : np.ndarray
            Instance labelled mask. Shape (H, W).
        n_rays : int, default=32
            Number of rays.
        n_segments : int, default=15
            Number of line segments the contour is divided into.
            The more segments used, the more detail is preserved with
            performance tradeoff.

    Returns
    -------
        np.ndarray:
            The radial distance maps. Shape (n_rays, H, W). Dtype: float64
    """
    eps = 1e-5
    H, W = inst_map.shape
    dist = np.zeros((n_rays, H, W), np.float32)

    inst_map = remove_small_objects(inst_map, min_size=5, out=inst_map)

    inst_list = list(np.unique(inst_map))
    if 0 in inst_list:
        inst_list.remove(0)

    for inst_id in inst_list:
        inst = np.array(inst_map == inst_id, np.int32)

        # major axis length of the object for padding
        props = regionprops(inst)
        pad = int(props[0].major_axis_length)

        sy, sx = ndi.find_objects(inst)[0]
        inst = np.pad(inst[sy, sx], pad)
        h, w = inst.shape
        y, x = np.nonzero(inst)

        # Evenly spaced point along the contour of the object.
        # The consecutive points are used as line segments
        yi, xi = evenly_spaced_points(dilation(inst, disk(1)), n_segments)
        yj = list(yi[1:])
        xj = list(xi[1:])
        yj.append(yi[0])
        xj.append(xi[0])
        points1 = np.array([yi, xi]).T
        points2 = np.array([yj, xj]).T

        ray_origins = np.array([y, x]).T
        st_rays = np.float32((2 * np.pi) / n_rays)
        for k in range(n_rays):
            theta = k * st_rays + 1e-4  # straight/right angles behave oddly (+ 1e-4)
            ray_directions = np.array([np.sin(theta), np.cos(theta)]).T

            # rays for each pixel with angle theta
            rays = (pad * 1.5) * ray_directions + ray_origins

            # Compute intersecting point for each contour line segment and each ray

            # ray endpoints
            y1 = ray_origins[:, 0]
            x1 = ray_origins[:, 1]
            y2 = rays[:, 0]
            x2 = rays[:, 1]

            dst = np.zeros((n_segments, h * w), dtype=np.float32)
            for t, (p1, p2) in enumerate(zip(points1, points2)):
                # line segments end points
                y3 = p1[0]
                x3 = p1[1]
                y4 = p2[0]
                x4 = p2[1]

                # find intersections
                denom = (x1 - x2) * (y3 - y4) - (y1 - y2) * (x3 - x4) + 1e-10
                det1 = x1 * y2 - y1 * x2
                det2 = x3 * y4 - y3 * x4

                px = (det1 * (x3 - x4) - (x1 - x2) * det2) / denom
                py = (det1 * (y3 - y4) - (y1 - y2) * det2) / denom

                # Set non-intersecting points to NaN
                px[(px - x1) * (px - x2) > eps] = np.nan
                px[(px - x3) * (px - x4) > eps] = np.nan
                py[(py - y1) * (py - y2) > eps] = np.nan
                py[(py - y3) * (py - y4) > eps] = np.nan

                d = ((y - py)) ** 2 + ((x - px)) ** 2
                d = np.sqrt(d) - 1.1  # crude correction for crossing boundaries

                dst[t, y * w + x] = d

            ray_dists = np.nansum(dst, axis=0)
            dist[k, sy.start + y - pad, sx.start + x - pad] = ray_dists[y * w + x]

    return dist



import torch
import torch.nn.functional as F

from ...utils import tensor_one_hot
from ..weighted_base_loss import WeightedBaseLoss


class FocalLoss(WeightedBaseLoss):
    def __init__(
        self,
        alpha: float = 0.5,
        gamma: float = 2.0,
        apply_sd: bool = False,
        apply_ls: bool = False,
        apply_svls: bool = False,
        edge_weight: float = None,
        class_weights: torch.Tensor = None,
        **kwargs
    ) -> None:
        """Focal loss.

        https://arxiv.org/abs/1708.02002

        Optionally applies, label smoothing, spatially varying label smoothing or
        weights at the object edges or class weights to the loss.

        Parameters
        ----------
            alpha : float, default=0.5
                Weight factor b/w [0,1].
            gamma : float, default=2.0
                Focusing factor.
            apply_sd : bool, default=False
                If True, Spectral decoupling regularization will be applied  to the
                loss matrix.
            apply_ls : bool, default=False
                If True, Label smoothing will be applied to the target.
            apply_svls : bool, default=False
                If True, spatially varying label smoothing will be applied to the target
            edge_weight : float, default=none
                Weight that is added to object borders.
            class_weights : torch.Tensor, default=None
                Class weights. A tensor of shape (n_classes,).
        """
        super().__init__(apply_sd, apply_ls, apply_svls, class_weights, edge_weight)
        self.alpha = alpha
        self.gamma = gamma
        self.eps = 1e-8

    def forward(
        self,
        yhat: torch.Tensor,
        target: torch.Tensor,
        target_weight: torch.Tensor = None,
        **kwargs
    ) -> torch.Tensor:
        """Compute the focal loss.

        Parameters
        ----------
            yhat : torch.Tensor
                The prediction map. Shape (B, C, H, W).
            target : torch.Tensor
                the ground truth annotations. Shape (B, H, W).
            target_weight : torch.Tensor, default=None
                The edge weight map. Shape (B, H, W).

        Returns
        -------
            torch.Tensor:
                Computed Focal loss (scalar).
        """
        input_soft = F.softmax(yhat, dim=1) + self.eps  # (B, C, H, W)
        num_classes = yhat.shape[1]
        target_one_hot = tensor_one_hot(target, num_classes)  # (B, C, H, W)
        assert target_one_hot.shape == yhat.shape

        if self.apply_svls:
            target_one_hot = self.apply_svls_to_target(
                target_one_hot, num_classes, **kwargs
            )

        if self.apply_ls:
            target_one_hot = self.apply_ls_to_target(
                target_one_hot, num_classes, **kwargs
            )

        weight = (1.0 - input_soft) ** self.gamma
        focal = self.alpha * weight * torch.log(input_soft)
        focal = target_one_hot * focal

        loss = -torch.sum(focal, dim=1)  # to (B, H, W)

        if self.apply_sd:
            loss = self.apply_spectral_decouple(loss, yhat)

        if self.class_weights is not None:
            loss = self.apply_class_weights(loss, target)

        if self.edge_weight is not None:
            loss = self.apply_edge_weights(loss, target_weight)

        return loss.mean()



from typing import Any, Dict, Tuple

__all__ = ["_create_stardist_args"]


def _create_stardist_args(
    depth: int,
    norm: str,
    act: str,
    conv: str,
    att: str,
    preact: bool,
    preattend: bool,
    short_skip: str,
    use_style: bool,
    block_type: str,
    merge_policy: str,
    skip_params: Dict[str, Any],
) -> Tuple[Dict[str, Any], ...]:
    """Create the args to build CellPose-Unet architecture."""
    skip_params = skip_params if skip_params is not None else {"k": None}

    return (
        {
            "merge_policy": merge_policy,
            "short_skips": (short_skip,),
            "block_types": ((block_type, block_type),),
            "kernel_sizes": ((3, 3),),
            "expand_ratios": ((1.0, 1.0),),
            "groups": ((1, 1),),
            "biases": ((False, False),),
            "normalizations": ((norm, norm),),
            "activations": ((act, act),),
            "convolutions": ((conv, conv),),
            "attentions": ((att, att),),
            "preactivates": ((preact, preact),),
            "preattends": ((preattend, preattend),),
            "use_styles": ((use_style, False),),
            "skip_params": {
                "short_skips": (short_skip,),
                "block_types": ((block_type,),),
                "convolutions": ((conv,),),
                "normalizations": ((norm,),),
                "activations": ((act,),),
                **skip_params,
            },
        },
    ) * depth



import re
from functools import partial
from pathlib import Path
from typing import Any, Dict, List, Tuple, Union

import cv2
import numpy as np
import scipy.io as sio

from .mask_utils import (
    bounding_box,
    fix_duplicates,
    get_inst_centroid,
    get_inst_types,
    label_semantic,
)
from .multiproc import run_pool

try:
    import tables as tb

    _has_tb = True
except ModuleNotFoundError:
    _has_tb = False


class FileHandler:
    """Class for handling file I/O."""

    @staticmethod
    def read_img(path: Union[str, Path]) -> np.ndarray:
        """Read an image & convert from bgr to rgb. (cv2 reads imgs in bgr).

        Parameters
        ----------
            path : str or Path
                Path to the image file.

        Returns
        -------
            np.ndarray:
                The image. Shape (H, W, 3).
        """
        path = Path(path)
        return cv2.cvtColor(cv2.imread(path.as_posix()), cv2.COLOR_BGR2RGB)

    @staticmethod
    def write_img(path: Union[str, Path], img: np.ndarray) -> None:
        """Write an image.

        Parameters
        ----------
            path : str or Path
                Path to the image file.
            img : np.ndarray
                The image to be written.

        """
        path = Path(path)
        cv2.imwrite(path.as_posix(), cv2.cvtColor(img, cv2.COLOR_RGB2BGR))

    @staticmethod
    def read_mat(
        path: Union[str, Path],
        key: str = "inst_map",
        retype: bool = True,
        return_all: bool = False,
    ) -> Union[np.ndarray, Dict[str, np.ndarray], None]:
        """Read a mask from a .mat file.

        If a mask is not found, return None

        Parameters
        ----------
            path : str or Path
                Path to the .mat file.
            key : str, default="inst_map"
                Name/key of the mask type that is being read from .mat
            retype : bool, default=True
                Convert the matrix type.
            return_all : bool, default=False
                Return the whole dict. Overrides the `key` arg.


        Raises
        ------
            ValueError: If an illegal key is given.

        Returns
        -------
            Union[np.ndarray, List[np.ndarray], None]:
                if return_all == False:
                    The instance/type/semantic labelled mask. Shape: (H, W).
                if return_all == True:
                    All the masks in the .mat file returned in a dictionary.
        """
        dtypes = {
            "inst_map": "int32",
            "type_map": "int32",
            "sem_map": "int32",
            "inst_centroid": "float64",
            "inst_type": "int32",
        }

        path = Path(path)
        if not path.exists():
            raise ValueError(f"{path} not found")

        try:
            mask = sio.loadmat(path.as_posix())
        except Exception:
            mask = None

        if not return_all:
            allowed = ("inst_map", "type_map", "inst_centroid", "inst_type", "sem_map")
            if key not in allowed:
                raise ValueError(f"Illegal key given. Got {key}. Allowed: {allowed}")

            try:
                mask = mask[key]
                if retype:
                    mask = mask.astype(dtypes[key])
            except Exception:
                mask = None

        return mask

    @staticmethod
    def get_geo_obj(what: str = "qupath") -> Dict[str, str]:
        """Get the dict format for a geojson obj.

        For example: get the obj in QuPath PathCellDetection obj

        Parameters
        ----------
            what : str
                One of "qupath", "simple"

        Returns
        -------
            Dict[str, Any]:
                A dictionary in geojson format.
        """
        allowed = ("qupath",)
        if what not in allowed:
            raise ValueError(f"Illegal `what`-arg. Got: {what}. Allowed: {allowed}")

        geo_obj = {}
        if what == "qupath":
            geo_obj.setdefault("type", "Feature")

            # PathCellAnnotation, PathCellDetection, PathDetectionObject
            geo_obj.setdefault("id", "PathCellDetection")
            geo_obj.setdefault("geometry", {"type": "Polygon", "coordinates": None})
            geo_obj.setdefault(
                "properties",
                {
                    "isLocked": "false",
                    "measurements": [],
                    "classification": {"name": None},
                },
            )

        return geo_obj

    @staticmethod
    def get_gson(
        inst: np.ndarray,
        type: np.ndarray,
        classes: Dict[str, int],
        x_offset: int = 0,
        y_offset: int = 0,
        geo_format: str = "qupath",
    ) -> Dict[str, Any]:
        """Get the labels in geojson format.

        Parameters
        ----------
            inst : np.ndarray
                Instance labelled mask. Shape: (H, W).
            type : np.ndarray
                Cell type labelled semantic segmentation mask. Shape: (H, W).
            classes : Dict[str, int]
                Class dict e.g. {"inflam":1, "epithelial":2, "connec":3}
            x_offset : int, default=0
                x-coordinate offset. (to set geojson to .mrxs wsi coordinates)
            y_offset : int, default=0
                y-coordinate offset. (to set geojson to .mrxs wsi coordinates)
            geo_format : str, default="qupath"
                The format for the geo object. "qupath" format allows the result file
                to be read with QuPath. "simple" format allows for geopandas etc.

        Returns
        -------
            Dict[str, Any]:
                A geojson dictionary of the instance labelled mask.
        """
        inst_map = fix_duplicates(inst)
        inst_list = list(np.unique(inst_map))
        if 0 in inst_list:
            inst_list.remove(0)

        geo_objs = []
        for inst_id in inst_list:
            # set up the annotation geojson obj

            # Get cell instance and cell type
            inst = np.array(inst_map == inst_id, np.uint8)
            inst_type = type[inst_map == inst_id].astype("uint8")
            inst_type = np.unique(inst_type)[0]

            inst_type = [key for key in classes.keys() if classes[key] == inst_type][0]

            # get the cell contour coordinates
            contours, _ = cv2.findContours(inst, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)

            # got a line instead of a polygon
            if contours[0].shape[0] < 3:
                continue

            # shift coordinates based on the offsets
            if x_offset:
                contours[0][..., 0] += x_offset
            if y_offset:
                contours[0][..., 1] += y_offset

            # Get the geojson obj
            geo_obj = FileHandler.get_geo_obj(what=geo_format)
            poly = contours[0].squeeze().tolist()
            poly.append(poly[0])  # close the polygon
            geo_obj["geometry"]["coordinates"] = [poly]
            geo_obj["properties"]["classification"]["name"] = inst_type
            geo_objs.append(geo_obj)

        return geo_objs

    @staticmethod
    def read_h5_patch(
        path: Union[Path, str],
        ix: int,
        return_im: bool = True,
        return_inst: bool = True,
        return_type: bool = True,
        return_sem: bool = False,
        return_name: bool = False,
        return_nitems: bool = False,
        return_all_names: bool = False,
    ) -> Dict[str, np.ndarray]:
        """Read img & mask patches at index `ix` from a hdf5 db.

        Parameters
        ----------
            path : Path or str
                Path to the h5-db.
            ix : int
                Index for the hdf5 db-arrays.
            return_im : bool, default=True
                If True, returns an image. (If the db contains these.)
            return_inst : bool, default=True
                If True, returns a instance labelled mask. (If the db contains these.)
            return_type : bool, default=True
                If True, returns a type mask. (If the db contains these.)
            return_sem : bool, default=False
                If True, returns a semantic mask, (If the db contains these.)
            return_name : bool, default=False
                If True, returns a name for the patch, (If the db contains these.)
            return_nitems : bool, default=False
                If True, returns the number of items in the db.
            return_all_names : bool, default=False
                If True, returns all the names in the db.

        Returns
        -------
            Dict[str, np.ndarray]:
                A Dict of numpy matrices. Img shape: (H, W, 3), mask shapes: (H, W).
                keys of the dict are: "im", "inst", "type", "sem"

        Raises
        ------
            IOError: If a mask that does not exist in the db is being read.
        """
        if not _has_tb:
            raise ModuleNotFoundError(
                "`FileHandler.read_h5_patch` method requires pytables library. "
                "Install with `pip install tables`."
            )

        path = Path(path)
        with tb.open_file(path.as_posix(), "r") as h5:
            out = {}

            if return_im:
                try:
                    out["image"] = h5.root.imgs[ix, ...]
                except Exception:
                    raise IOError(
                        "The HDF5 database does not contain images. Try "
                        "setting `return_im=False`"
                    )

            if return_inst:
                try:
                    out["inst"] = h5.root.insts[ix, ...]
                except Exception:
                    raise IOError(
                        "The HDF5 database does not contain instance labelled masks. "
                        "Try setting `return_inst=False`"
                    )

            if return_type:
                try:
                    out["type"] = h5.root.types[ix, ...]
                except Exception:
                    raise IOError(
                        "The HDF5 database does not contain type masks. Try setting "
                        "`return_type = False` "
                    )

            if return_sem:
                try:
                    out["sem"] = h5.root.areas[ix, ...]
                except Exception:
                    raise IOError(
                        "The HDF5 database does not contain semantic masks. Try "
                        "setting `return_sem = False`"
                    )

            if return_name:
                try:
                    fn = h5.root.fnames[ix]
                    out["name"] = Path(fn.decode("UTF-8"))
                except Exception:
                    raise IOError(
                        "The HDF5 database does not contain patch names. Try "
                        "setting `return_name = False`"
                    )

            if return_nitems:
                try:
                    out["nitems"] = h5.root._v_attrs.n_items
                except Exception:
                    raise IOError(
                        "The HDF5 database does not contain attribute ``nitems. Try "
                        "setting `return_nitems = False`"
                    )

            if return_all_names:
                try:
                    names = h5.root.fnames[:]
                    out["names"] = [Path(n.decode("UTF-8")) for n in names]
                except Exception:
                    raise IOError(
                        "The HDF5 database does not contain patch names. Try "
                        "setting `return_all_names = False`"
                    )

            return out

    @staticmethod
    def write_mat(
        fname: Union[str, Path],
        inst: np.ndarray,
        type: np.ndarray = None,
        sem: np.ndarray = None,
        compute_centorids: bool = False,
        compute_bboxes: bool = False,
    ) -> None:
        """
        Write multiple masks to .mat file.

        Keys always present in the file: "inst_map", "inst_type"

        Optional keys: "type_map", "sem_map", "inst_bbox", "inst_centroid"

        Parameters
        ----------
            fname : str or Path
                The file name of the .mat file.
            inst : np.ndarray
                Instance labelled mask. Shape: (H, W).
            type : np.ndarray
                Cell type labelled semantic segmentation mask. Shape: (H, W).
            sem : np.ndarray
                Tissue type labelled semantic segmentation mask. Shape: (H, W).
            compute_centroids : bool, optional
                Flag to tompute instance centorids.
            compute_bboxes : bool, optional
                Flag to tompute instance bboxes.
        """
        fname = Path(fname)
        if not fname.parent.exists():
            raise ValueError(
                f"The directory: {fname.parent.as_posix()} does not exist."
            )

        inst_map = fix_duplicates(inst)
        inst_types = get_inst_types(inst, type)

        res = {
            "inst_map": inst_map,
            "inst_type": inst_types,
        }

        if compute_centorids:
            centroids = get_inst_centroid(inst_map)
            res["inst_centroid"] = centroids

        if compute_bboxes:
            inst_ids = list(np.unique(inst_map)[1:])
            bboxes = np.array(
                [bounding_box(np.array(inst_map == id_, np.uint8)) for id_ in inst_ids]
            )
            res["inst_bbox"] = bboxes

        if type is not None:
            res["type_map"] = type

        if sem is not None:
            res["sem_map"] = sem

        sio.savemat(
            file_name=fname.with_suffix(".mat").as_posix(),
            mdict=res,
        )

    @staticmethod
    def write_gson(
        fname: Union[str, Path],
        inst: np.ndarray,
        type: np.ndarray = None,
        classes: Dict[str, int] = None,
        x_offset: int = 0,
        y_offset: int = 0,
        geo_format: str = "qupath",
    ) -> None:
        """Convert the instance labelled mask into geojson obj or write it .json file.

        Parameters
        ----------
            fname : str | Path, optional
                File name for the .json file.
            inst : np.ndarray
                Instance labelled mask. Shape: (H, W).
            type : np.ndarray, optional
                Cell type labelled semantic segmentation mask. Shape: (H, W). If None,
                the classes of the objects will be set to {background: 0, foreground: 1}
            classes : Dict[str, int], optional
                Class dict e.g. {"inflam":1, "epithelial":2, "connec":3}. Ignored if
                `type` is None.
            x_offset : int, default=0
                x-coordinate offset. (to set geojson to .mrxs wsi coordinates)
            y_offset : int, default=0
                y-coordinate offset. (to set geojson to .mrxs wsi coordinates)
            geo_format : str, default="qupath"
                The format for the geo object. "qupath" format allows the result file
                to be read with QuPath. "simple" format allows for geopandas etc.

        Raises
        ------
            ModuleNotFoundError: If geojson is not installed.
            ValueError: If `classes` is set to None when `type` is given.

        Returns
        -------
            Dict[str, Any]:
                A dictionary with geojson fields.
        """
        try:
            import geojson
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "To use the `FileHandler.mask2geojson`, pytorch-lightning is required. "
                "Install with `pip install pytorch-lightning`"
            )

        fname = Path(fname)
        if not fname.parent.exists():
            raise ValueError(
                f"The directory: {fname.parent.as_posix()} does not exist."
            )

        if type is None:
            type = inst > 0
            classes = {"background": 0, "foreground": 1}
        else:
            if classes is None:
                raise ValueError(
                    "`classes` cannot be None if `type` semgentation map is given."
                )

        geo_objs = FileHandler.get_gson(
            inst, type, classes, x_offset, y_offset, geo_format
        )

        fname = fname.with_suffix(".json")

        with fname.open("w") as out:
            geojson.dump(geo_objs, out)

    @staticmethod
    def save_masks(
        fname: str,
        maps: Dict[str, np.ndarray],
        format: str = ".mat",
        json_format: str = "qupath",
        classes_type: Dict[str, str] = None,
        classes_sem: Dict[str, str] = None,
        offsets: bool = False,
        **kwargs,
    ) -> None:
        """Save model outputs to .mat or geojson .json file.

        NOTE: If .json format is used, two files are written if both inst_map and
        sem_map are given. The sem_map .json has a suffix '{}_areas.json' and
        the inst_map .json has suffix '{}_cells.json'

        Parameters
        ----------
            fname : str
                Name for the output-file.
            maps : Dict[str, np.ndarray]
                model output names mapped to model outputs.
                E.g. {"sem": np.ndarray, "type": np.ndarray, "inst": np.ndarray}
            format : str
                One of ".mat" or ".json"
            json_format : str, default="qupath"
                The geojson format. One of "qupath", "simple". Ignored if format is not
                ".json".
            classes_type : Dict[str, str], optional
                Cell type dictionary. e.g. {"inflam":1, "epithelial":2, "connec":3}.
                This is required only if `format` == `json`.
            classes_sem : Dict[str, str], otional
                Tissue type dictionary. e.g. {"tissue1":1, "tissue2":2, "tissue3":3}
                This is required only if `format` == `json`.
            offsets : bool, default=False
                If True, geojson coords are shifted by the offsets that are encoded in
                the filenames (e.g. "x-1000_y-4000.png"). Ignored if `format` != ".json"
        """
        fname = Path(fname)
        allowed = (".mat", ".json")
        if format not in allowed:
            raise ValueError(
                f"Illegal file-format. Got: {format}. Allowed formats: {allowed}"
            )

        if format == ".mat":
            FileHandler.write_mat(fname, **maps)
        elif format == ".json":
            offs = FileHandler.get_offset(fname.name) if offsets else {"x": 0, "y": 0}

            if "inst" in maps.keys():
                type_map = None
                if "type" in maps.keys():
                    type_map = maps["type"]

                # Create directory for the cell seg results
                save_dir = fname.parent / "cells"
                if not Path(save_dir).exists():
                    Path(save_dir).mkdir(parents=True, exist_ok=True)

                fn = save_dir / f"{fname.name}_cells"
                FileHandler.write_gson(
                    fname=fn,
                    inst=maps["inst"],
                    type=type_map,
                    classes=classes_type,
                    geo_format=json_format,
                    x_offset=offs["x"],
                    y_offset=offs["y"],
                )
            if "sem" in maps.keys():
                if classes_sem is None:
                    raise ValueError(
                        "When saving to .json `classes_sem` can't be None, "
                        "if the output masks contains tissue type predictions."
                    )

                # Create directory for the area seg results
                save_dir = fname.parent / "areas"
                if not Path(save_dir).exists():
                    Path(save_dir).mkdir(parents=True, exist_ok=True)

                fn = save_dir / f"{fname.name}_areas"

                FileHandler.write_gson(
                    fname=fn,
                    inst=label_semantic(maps["sem"]),
                    type=maps["sem"],
                    classes=classes_sem,
                    geo_format=json_format,
                    x_offset=offs["x"],
                    y_offset=offs["y"],
                )

    @staticmethod
    def save_masks_parallel(
        maps: List[Dict[str, np.ndarray]],
        fnames: List[str],
        format: str = ".mat",
        geo_format="qupath",
        classes_type: Dict[str, str] = None,
        classes_sem: Dict[str, str] = None,
        offsets: bool = False,
        pooltype: str = "thread",
        maptype: str = "amap",
        **kwargs,
    ) -> None:
        """Save the model output masks to a folder. (multi-threaded).

        Parameters
        ----------
            maps : List[Dict[str, np.ndarray]]
                The model output map dictionaries in a list.
            fnames : List[str]
                Name for the output-files. (In the same order with `maps`)
            format : str
                One of ".mat" or ".json"
            geo_format : str, default="qupath"
                The geojson format. One of "qupath", "simple". Ignored if format is not
                ".json".
            classes_type : Dict[str, str], optional
                Cell type dictionary. e.g. {"inflam":1, "epithelial":2, "connec":3}.
                This is required only if `format` == `json`.
            classes_sem : Dict[str, str], otional
                Tissue type dictionary. e.g. {"tissue1":1, "tissue2":2, "tissue3":3}
                This is required only if `format` == `json`.
            offsets : bool, default=False
                If True, geojson coords are shifted by the offsets that are encoded in
                the filenames (e.g. "x-1000_y-4000.png"). Ignored if `format` != ".json"
            pooltype : str, default="thread"
                The pathos pooltype. Allowed: ("process", "thread", "serial").
                Defaults to "thread". (Fastest in benchmarks.)
            maptype : str, default="amap"
                The map type of the pathos Pool object.
                Allowed: ("map", "amap", "imap", "uimap")
                Defaults to "amap". (Fastest in benchmarks).
        """
        func = partial(
            FileHandler._save_masks,
            format=format,
            geo_format=geo_format,
            classes_type=classes_type,
            classes_sem=classes_sem,
            offsets=offsets,
        )

        args = tuple(zip(fnames, maps))
        run_pool(func, args, ret=False, pooltype=pooltype, maptype=maptype)

    @staticmethod
    def _save_masks(
        args: Tuple[str, Dict[str, np.ndarray]],
        format: str,
        geo_format: str,
        classes_type: Dict[str, str],
        classes_sem: Dict[str, str],
        offsets: bool,
    ) -> None:
        """Unpacks the args for `save_mask` to enable multi-threading."""
        return FileHandler.save_masks(
            *args,
            format=format,
            geo_format=geo_format,
            classes_type=classes_type,
            classes_sem=classes_sem,
            offsets=offsets,
        )

    @staticmethod
    def get_split(string: str) -> List[str]:
        """Try splitting a coord-string with "-" and "_" on a string."""
        xy = string.split("-")
        if len(xy) > 1:
            return xy
        xy = string.split("_")
        if len(xy) > 1:
            return xy
        else:
            return list(filter(None, re.split(r"(\d+)", string)))

    @staticmethod
    def get_offset(fname: str) -> Dict[str, int]:
        """Get the offsets.

        I.e. If a filename contains x- and y- coordinates, return them.
        """
        coords = re.findall(r"([xy][ -_]\d+)", fname)
        offsets = {}
        for coord in coords:
            xy = FileHandler.get_split(coord)
            offsets[xy[0]] = int(xy[1])

        return offsets



from typing import Any, Dict, List, Optional, Tuple, Union

import torch
import torch.nn as nn

from ..modules import ChannelPool, ConvLayer, Transformer2D, Up
from .long_skips import LongSkip

__all__ = ["DecoderStage"]


class DecoderStage(nn.Module):
    def __init__(
        self,
        stage_ix: int,
        dec_channels: Tuple[int, ...],
        dec_dims: Tuple[int, ...],
        skip_channels: Tuple[int, ...],
        long_skip: str = "unet",
        merge_policy: str = "sum",
        skip_params: Optional[Dict[str, Any]] = None,
        upsampling: str = "fixed-unpool",
        n_conv_layers: Optional[int] = 1,
        style_channels: Optional[int] = None,
        layer_residual: Optional[bool] = False,
        n_conv_blocks: Optional[Tuple[int, ...]] = (2,),
        short_skips: Optional[Tuple[str, ...]] = ("residual",),
        expand_ratios: Optional[Tuple[float, float]] = ((1.0, 1.0),),
        block_types: Optional[Tuple[Tuple[str, ...], ...]] = (("basic", "basic"),),
        normalizations: Optional[Tuple[Tuple[str, ...], ...]] = (("bn", "bn"),),
        activations: Optional[Tuple[Tuple[str, ...], ...]] = (("relu", "relu"),),
        convolutions: Optional[Tuple[Tuple[str, ...], ...]] = (("conv", "conv"),),
        attentions: Optional[Tuple[Tuple[str, ...], ...]] = ((None, "se"),),
        preactivates: Optional[Tuple[Tuple[bool, ...], ...]] = ((False, False),),
        preattends: Optional[Tuple[Tuple[bool, ...], ...]] = ((False, False),),
        use_styles: Optional[Tuple[Tuple[bool, ...], ...]] = ((False, False),),
        kernel_sizes: Optional[Tuple[Tuple[int, ...]]] = ((3, 3),),
        groups: Optional[Tuple[Tuple[int, ...]]] = ((1, 1),),
        biases: Optional[Tuple[Tuple[bool, ...]]] = ((False, False),),
        n_transformers: Optional[int] = None,
        n_transformer_blocks: Optional[Tuple[int, ...]] = (1,),
        transformer_blocks: Optional[Tuple[Tuple[str, ...], ...]] = (("exact",),),
        transformer_computations: Optional[Tuple[Tuple[str, ...], ...]] = (("basic",),),
        transformer_biases: Optional[Tuple[Tuple[bool, ...], ...]] = ((False,),),
        transformer_dropouts: Optional[Tuple[Tuple[float, ...], ...]] = ((0.0,),),
        transformer_layer_scales: Optional[Tuple[Tuple[bool, ...], ...]] = ((False,),),
        transformer_params: Optional[List[Dict[str, Any]]] = None,
        **kwargs,
    ) -> None:
        """Build a decoder stage.

        Operations in each decoder stage:
        1. Upsample.
        2. Long skip from encoder to decoder if not the last stage.
        3. Conv block (optional, applied by default).
        4. Transformer block (optional, not applied by default).

        Parameters
        ----------
            stage_ix : int
                The index number of the current decoder stage.
            dec_channels : Tuple[int, ...]
                The number of output channels in the decoder output stages. First elem
                is the number of channels in the encoder head or bottleneck.
            dec_dims : Tuple[int, ...]
                Tuple of the heights/widths of each encoder/decoder feature map
                e.g. (8, 16, 32, 64, 128, 256). Feature maps are assumed to be square.
            skip_channels : Tuple[int, ...]
                List of the number of channels in the encoder skip tensors. Ignored if
                `long_skip` == None.
            long_skip : str, default="unet"
                long skip method to be used.
                Allowed: "cross-attn", "unet", "unetpp", "unet3p", "unet3p-lite", None
            merge_policy : str, default="sum"
                The long skip merge policy. One of: "sum", "cat"
            skip_params : Optional[Dict]
                Extra keyword arguments for the skip-connection module. These depend
                on the skip module. Refer to specific skip modules for more info.
            upsampling : str, default="fixed-unpool"
                Name of the upsampling method.
            n_conv_layers : int, default=1
                The number of conv layers inside one decoder stage.
            style_channels : int, default=None
                Number of style vector channels. If None, style vectors are ignored.
                Also, ignored if `n_conv_layers` is None.
            layer_residual : bool, optional, default=False
                Apply a layer level residual short skip at each layer. I.e x + layer(x).
                Ignored if `n_conv_layers` is None.
            n_conv_blocks : Tuple[int, ...], default=(2,)
                Number of conv-blocks inside each conv layer. The tuple-length has to
                match `n_conv_layers`. Ignored if `n_conv_layers` is None.
            short_skips : str, optional,  default=("residual", )
                The short skip methods used inside the conv layers. Ignored if
                `n_conv_layers` is None.
            expand_ratios : Tuple[float, ...], default=((1.0, 1.0),):
                Expansion/Squeeze ratios for the out channels of each conv block.
                The tuple-length has to match `n_conv_layers`. Ignored if
                `n_conv_layers` is None.
            block_types : Tuple[Tuple[str, ...], ...], default=(("basic", "basic"), )
                The type of the convolution blocks in the conv blocks inside the layers.
                The tuple-length has to match `n_conv_layers`. Ignored if
                `n_conv_layers` is None.
            normalizations : Tuple[Tuple[str, ...], ...], default: (("bn", "bn"), )
                Normalization methods used in the conv blocks inside the conv layers.
                The tuple-length has to match `n_conv_layers`. Ignored if
                `n_conv_layers` is None.
            activations : Tuple[Tuple[str, ...], ...], default: (("relu", "relu"), )
                Activation methods used inside the conv layers.
                The tuple-length has to match `n_conv_layers`. Ignored if
                `n_conv_layers` is None.
            attentions : Tuple[Tuple[str, ...], ...], default: ((None, "se"), )
                Channel-attention methods used inside the conv layers.
                The tuple-length has to match `n_conv_layers`. Ignored if
                `n_conv_layers` is None. Allowed methods.: "se", "scse", "gc", "eca",
                "msca", None.
            preactivates Tuple[Tuple[bool, ...], ...], default: ((False, False), )
                Boolean flags for the conv layers to use pre-activation.
                The tuple-length has to match `n_conv_layers`. Ignored if
                `n_conv_layers` is None.
            preattends Tuple[Tuple[bool, ...], ...], default: ((False, False), )
                Boolean flags for the conv layers to use pre-activation.
                The tuple-length has to match `n_conv_layers`. Ignored if
                `n_conv_layers` is None.
            use_styles : Tuple[Tuple[bool, ...], ...], default=((False, False), )
                Boolean flags for the conv layers to add style vectors at each block.
                The tuple-length has to match `n_conv_layers`. Ignored if
                `n_conv_layers` is None.
            kernel_sizes : Tuple[int, ...], default=((3, 3),)
                The size of the convolution kernels in each conv block.
                The tuple-length has to match `n_conv_layers`. Ignored if
                `n_conv_layers` is None.
            groups : int, default=((1, 1),)
                Number of groups for the kernels in each convolution blocks.
                The tuple-length has to match `n_conv_layers`. Ignored if
                `n_conv_layers` is None.
            biases : bool, default=((False, False),)
                Include bias terms in the convolution blocks.
                The tuple-length has to match `n_conv_layers`. Ignored if
                `n_conv_layers` is None.
            n_transformers : int, optional
                Number of self-attention tranformers applied after the conv-layer.
                If this is None, no transformers will be added.
            n_transformer_blocks : int, default=(2, ), optional
                Number of multi-head self attention blocks used in the transformer
                layers. Ignored if `n_transformers` is None.
            transformer_blocks : Tuple[Tuple[str, ...], ...], default=(("basic",),)
                The name of the SelfAttentionBlocks in the TransformerLayer(s).
                Allowed values: "basic", "slice", "flash". Ignored if `n_transformers`
                is None. Length of the tuple has to equal `n_transformer_blocks`.
                Allowed names: ("exact", "linformer").
            transformer_computations : Tuple[Tuple[str, ...],...], default=(("basic",),)
                The way of computing the attention matrices in the SelfAttentionBlocks
                in the TransformerLayer(s). Length of the tuple has to equal
                `n_transformer_blocks`. Allowed styles: "basic". "slice", "flash",
                "memeff", "slice-memeff".
            transformer_biases : Tuple[Tuple[bool, ...], ...], default=((False,),)
                Flags, whether to use biases in the transformer layers. Ignored if
                `n_transformers` is None.
            transformer_dropoouts : Tuple[Tuple[float, ...], ...], default=((0.0,),)
                Dropout probabilities in the transformer layers. Ignored if
                `n_transformers` is None.
            transformer_layer_scales : Tuple[Tuple[bool, ...], ...], default=((False,),)
                Flags, whether to use layer scales in the transformer layers. Ignored if
                `n_transformers` is None.
            transformer_params : List[Dict[str, Any]]
                Extra keyword arguments for the transformer layers. Refer to
                `Transformer2D` module for more info. Ignored if `n_transformers`
                is None.

        Raises
        ------
            ValueError:
                If lengths of the conv layer tuple args are not equal to `n_conv_layers`
                If lengths of the transformer layer tuple args are not equal to
                `n_transformers`.
        """
        super().__init__()

        self.n_conv_layers = n_conv_layers
        self.n_transformers = n_transformers
        self.long_skip = long_skip
        self.stage_ix = stage_ix
        self.in_channels = dec_channels[stage_ix]
        self.out_channels = dec_channels[stage_ix + 1]

        # upsampling method
        self.upsample = Up(upsampling)

        # long skip connection method
        self.skip = LongSkip(
            name=long_skip,
            merge_policy=merge_policy,
            stage_ix=self.stage_ix,
            in_channels=self.in_channels,
            dec_channels=dec_channels,
            skip_channels=skip_channels,
            dec_dims=dec_dims,
            **skip_params if skip_params is not None else {"k": None},
        )

        # Set up n layers of conv blocks
        layer = None  # placeholder
        if n_conv_layers is not None:

            # check that the conv-layer tuple-args are not illegal.
            self._check_tuple_args(
                "conv-layer related",
                "n_conv_layers",
                n_conv_layers,
                all_args=locals(),
                skip_args=(
                    skip_channels,
                    dec_channels,
                    dec_dims,
                    transformer_blocks,
                    transformer_computations,
                    n_transformer_blocks,
                    transformer_biases,
                    transformer_dropouts,
                    transformer_layer_scales,
                ),
            )

            # set up the conv-layers.
            self.conv_layers = nn.ModuleDict()
            for i in range(n_conv_layers):
                n_in_feats = self.skip.out_channels if i == 0 else layer.out_channels
                layer = ConvLayer(
                    in_channels=n_in_feats,
                    out_channels=self.out_channels,
                    n_blocks=n_conv_blocks[i],
                    layer_residual=layer_residual,
                    style_channels=style_channels,
                    short_skip=short_skips[i],
                    expand_ratios=expand_ratios[i],
                    block_types=block_types[i],
                    normalizations=normalizations[i],
                    activations=activations[i],
                    convolutions=convolutions[i],
                    preactivates=preactivates[i],
                    preattends=preattends[i],
                    use_styles=use_styles[i],
                    attentions=attentions[i],
                    kernel_sizes=kernel_sizes[i],
                    groups=groups[i],
                    biases=biases[i],
                    **kwargs,
                )
                self.conv_layers[f"conv_layer{i + 1}"] = layer

            self.out_channels = layer.out_channels

        # set in_channels for final operations
        in_channels = (
            self.skip.out_channels if n_conv_layers is None else self.out_channels
        )

        if n_transformers is not None:

            # check that the transformer-layer tuple args are not illegal.
            self._check_tuple_args(
                "transformer related",
                "n_transformers",
                n_transformers,
                all_args=locals(),
                skip_args=(
                    skip_channels,
                    dec_channels,
                    dec_dims,
                    n_conv_blocks,
                    short_skips,
                    expand_ratios,
                    block_types,
                    normalizations,
                    activations,
                    convolutions,
                    attentions,
                    preactivates,
                    preattends,
                    use_styles,
                    kernel_sizes,
                    groups,
                    biases,
                ),
            )

            # set up the transformer layers
            self.transformers = nn.ModuleDict()
            for i in range(n_transformers):
                tr = Transformer2D(
                    in_channels=in_channels,
                    n_blocks=n_transformer_blocks[i],
                    block_types=transformer_blocks[i],
                    computation_types=transformer_computations[i],
                    biases=transformer_biases[i],
                    dropouts=transformer_dropouts[i],
                    layer_scales=transformer_layer_scales[i],
                    **transformer_params
                    if transformer_params is not None
                    else {"k": None},
                )
                self.transformers[f"tr_layer_{i + 1}"] = tr

        # add a channel pooling layer at the end if no conv-layers are set up
        if n_conv_layers is None:
            self.ch_pool = ChannelPool(
                in_channels=in_channels,
                out_channels=self.out_channels,
                normalization=normalizations[0][0],
                convolution=convolutions[0][0],
            )

    def _check_tuple_args(
        self,
        case: str,
        var: str,
        n: int,
        all_args: Dict[str, Any],
        skip_args: Tuple[Any, ...],
    ) -> None:
        """Check for illegal tuple-args."""
        illegal_args = [
            (k, a)
            for k, a in all_args.items()
            if isinstance(a, tuple) and a not in skip_args and len(a) != n
        ]

        if illegal_args:
            raise ValueError(
                f"All {case} tuple-arg lengths need to be equal to `{var}`={n}. "
                f"Illegal args: {illegal_args}"
            )

    def forward(
        self,
        x: torch.Tensor,
        skips: Tuple[torch.Tensor, ...],
        extra_skips: Tuple[torch.Tensor, ...] = None,
        style: torch.Tensor = None,
    ) -> Tuple[torch.Tensor, Union[None, Tuple[torch.Tensor, ...]]]:
        """Forward pass of the decoder stage.

        Parameters
        ----------
            x : torch.Tensor
                Input tensor. Shape (B, C, H, W).
            skips : Tuple[torch.Tensor, ...]
                All of feature maps from consecutive encoder blocks.
                Order is bottom up. Shapes: (B, C, H, W).
            extra_skips : Tuple[torch.Tensor, ...], default=None
                Extra skip connections. Used in unet3+ and unet++.
            style : torch.Tensor
                Style vector. Shape (B, C).

        Returns
        -------
            Tuple[torch.Tensor, Union[None, Tuple[torch.Tensor, ...]]]:
                Output torch.Tensor and extra skip torch.Tensors. If no extra
                skips are present, returns None as the second return value.
        """
        x = self.upsample(x)  # (B, in_channels, H, W)

        # long skip (B, in_channels(+skip_channels), H, W)
        x = self.skip(x, ix=self.stage_ix, skips=skips, extra_skips=extra_skips)

        # unetpp returns extra skips
        extra_skips = x[1] if self.long_skip == "unetpp" else None
        x = x[0] if self.long_skip == "unetpp" else x

        # conv layers
        if self.n_conv_layers is not None:
            for conv_layer in self.conv_layers.values():
                x = conv_layer(x, style)  # (B, out_channels, H, W)

        # transformer layers
        if self.n_transformers is not None:
            for transformer in self.transformers.values():
                x = transformer(x)  # (B, long_skip_channels/out_channels, H, W)

        # channel pool if conv-layers are skipped.
        if self.n_conv_layers is None:
            x = self.ch_pool(x)  # (B, out_channels, H, W)

        return x, extra_skips



import pytest
import torch

from cellseg_models_pytorch.modules import ConvLayer


@pytest.mark.parametrize("convblocks", [("basic", "dws"), ("mbconv", "fmbconv")])
@pytest.mark.parametrize("short_skip", ["residual", "dense", "basic"])
def test_conv_block_forward(convblocks, short_skip):
    input = torch.rand([1, 8, 16, 16])
    layer = ConvLayer(
        in_channels=8,
        out_channels=4,
        block_types=convblocks,
        short_skip=short_skip,
        attentions=(None, "gc"),
        expand_ratio=1.0,
    )

    output = layer(input)

    assert output.shape == torch.Size([1, 4, 16, 16])
    assert output.dtype == input.dtype


@pytest.mark.parametrize("convblocks", [("basic", "dws"), ("mbconv", "fmbconv")])
@pytest.mark.parametrize("short_skip", ["residual", "dense", "basic"])
def test_conv_block_backward(convblocks, short_skip):
    input = torch.rand([1, 8, 16, 16])
    layer = ConvLayer(
        in_channels=8,
        out_channels=4,
        block_types=convblocks,
        short_skip=short_skip,
        attentions=(None, "gc"),
        expand_ratio=1.0,
    )

    output = layer(input)
    output.mean().backward()

    assert output.shape == torch.Size([1, 4, 16, 16])
    assert output.dtype == input.dtype



"""
Imported from the cellpose repo. Minor code style mods for readability.

Copyright © 2020 Howard Hughes Medical Institute

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer. Redistributions in binary
form must reproduce the above copyright notice, this list of conditions and
the following disclaimer in the documentation and/or other materials provided
with the distribution. Neither the name of HHMI nor the names of its contributors
may be used to endorse or promote products derived from this software without
specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS “AS IS”
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.
"""
import numpy as np
from scipy.ndimage import binary_fill_holes, find_objects
from skimage.color import hsv2rgb

from cellseg_models_pytorch.utils import percentile_normalize, percentile_normalize99

__all__ = ["gen_flows", "fill_holes_and_remove_small_masks", "normalize_field"]


def normalize_field(mu: np.ndarray) -> np.ndarray:
    """Normalize the flow field.

    Parameters
    ----------
        mu : np.ndarray
            The un-normalized y- and x- flows. Shape (2, H, W).

    Returns
    -------
        np.ndarray:
            The normalized y- and x- flows. Shape (2, H, W).
    """
    mag = np.sqrt(np.nansum(mu**2, axis=0))
    mask = np.logical_and(mag != 0, ~np.isnan(mag))
    out = np.zeros_like(mu)
    out = np.divide(mu, mag, out=out, where=mask)

    return out


def gen_flows(hover: np.ndarray) -> np.ndarray:
    """Convert Horizontal and Vertical gradients to cellpose flows.

    Parameters
    ----------
        flows : np.ndarray
            Horizontal and Vertical flows. Shape: (2, H, W).

    Returns
    -------
        np.ndarray:
            The optical flow representation. Shape (H, W, 3).
    """
    enhanced = percentile_normalize99(hover, amin=-1, amax=1)
    H = (np.arctan2(enhanced[0], enhanced[1]) + np.pi) / (2 * np.pi)
    S = percentile_normalize(enhanced[0] ** 2 + enhanced[1] ** 2)
    HSV = np.stack([H, S, S], axis=-1)
    HSV = np.clip(HSV, a_min=0.0, a_max=1.0)
    flow = (hsv2rgb(HSV) * 255).astype(np.uint8)

    return flow


def fill_holes_and_remove_small_masks(
    inst_map: np.ndarray, min_size: int = 15
) -> np.ndarray:
    """Fill holes in inst_map and discard objects smaller than `min_size`.

    Parameters
    ----------
        inst_map : np.ndarray
            Instance labelled mask. Shape (H, W).
        min_size : int, default=15
            Minimum number of pixels per mask, can turn off with -1.

    Returns
    -------
        np.ndarray:
            Processed iinstance labelled mask. Shape (H, W).

    Raises
    ------
        ValueError: If input has wrong shape.
    """
    if inst_map.ndim != 2:
        raise ValueError(f"`inst_map` shape need to be 2D. Got {inst_map.shape}.")

    j = 0
    slices = find_objects(inst_map)
    for i, slc in enumerate(slices):
        if slc is not None:

            msk = inst_map[slc] == (i + 1)
            npix = msk.sum()

            if min_size > 0 and npix < min_size:
                inst_map[slc][msk] = 0
            else:
                msk = binary_fill_holes(msk)
                inst_map[slc][msk] = j + 1
                j += 1

    return inst_map



from typing import Tuple

import torch
import torch.nn as nn

__all__ = ["CatBlock"]


class CatBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        skip_channels: Tuple[int, ...],
        **kwargs,
    ) -> None:
        """Merge skip blocks by concatenation.

        Parameters
        ----------
            in_channels : int
                Number of input channels.
            skip_channels : Tuple[int, ...]
                Number of skip channels.
        """
        super().__init__()
        self.in_channels = in_channels
        self.skip_channels = skip_channels

    @property
    def out_channels(self) -> int:
        """Out channels."""
        return sum([self.in_channels] + list(self.skip_channels))

    def forward(
        self,
        x: torch.Tensor,
        skips: Tuple[torch.Tensor, ...],
    ) -> torch.Tensor:
        """Forward of the cat block.

        Parameters
        ----------
            x : torch.Tensor
                Input tensor. Shape: (B, C, H, W)
            skips : Tuple[torch.Tensor, ...]
                All the skip features in a list. Shapes: (B, C, H, W).

        Returns
        -------
            torch.Tensor:
                The concatenated output tensor. Shape (B, C+skip_channels, H, W).
        """
        return torch.cat([x, *skips], dim=1)

    def extra_repr(self) -> str:
        """Add extra info to print."""
        s = "in_channels={in_channels}, skip_channels={skip_channels}"
        s = s.format(**self.__dict__) + f", out_channels={self.out_channels}"
        return s
